# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
A tool to test SMI handlers for pointer validation vulnerabilities

Reference:
    - Presented in CanSecWest 2015:
        - c7zero.info: `A New Class of Vulnerability in SMI Handlers of BIOS/UEFI Firmware <http://www.c7zero.info/stuff/ANewClassOfVulnInSMIHandlers_csw2015.pdf>`_


Usage:
``chipsec_main -m tools.smm.smm_ptr -l log.txt \``
``[-a <mode>,<config_file>|<smic_start:smic_end>,<size>,<address>]``

- ``mode``: SMI fuzzing mode

    * ``config`` = use SMI configuration file <config_file>
    * ``fuzz`` = fuzz all SMI handlers with code in the range <smic_start:smic_end>
    * ``fuzzmore`` = fuzz mode + pass 2nd-order pointers within buffer to SMI handlers
    * ``scan`` = fuzz mode + time measurement to identify SMIs that trigger long-running code paths
- ``size``: size of the memory buffer (in Hex)
- ``address``: physical address of memory buffer to pass in GP regs to SMI handlers (in Hex)

    * ``smram`` = option passes address of SMRAM base (system may hang in this mode!)

In ``config`` mode, SMI configuration file should have the following format

::

    SMI_code=<SMI code> or *
    SMI_data=<SMI data> or *
    RAX=<value of RAX> or * or PTR or VAL
    RBX=<value of RBX> or * or PTR or VAL
    RCX=<value of RCX> or * or PTR or VAL
    RDX=<value of RDX> or * or PTR or VAL
    RSI=<value of RSI> or * or PTR or VAL
    RDI=<value of RDI> or * or PTR or VAL
    [PTR_OFFSET=<offset to pointer in the buffer>]
    [SIG=<signature>]
    [SIG_OFFSET=<offset to signature in the buffer>]
    [Name=<SMI name>]
    [Desc=<SMI description>]

Where:

    - ``[]``: optional line
    - ``*``: Don't Care (the module will replace * with 0x0)
    - ``PTR``: Physical address SMI handler will write to (the module will replace PTR with physical address provided as a command-line argument)
    - ``VAL``: Value SMI handler will write to PTR address (the module will replace VAL with hardcoded _FILL_VALUE_xx)

Examples:

    >>> chipsec_main.py -m tools.smm.smm_ptr
    >>> chipsec_main.py -m tools.smm.smm_ptr -a fuzzmore,0x0:0xFF -l smm.log
    >>> chipsec_main.py -m tools.smm.smm_ptr -a scan,0x0:0xff

.. warning ::

    - This is a potentially destructive test

"""

import struct
import os
import sys
import math

from chipsec.module_common import BaseModule, SMM
from chipsec.library.returncode import ModuleResult
from chipsec.library.file import write_file
from chipsec.library.logger import print_buffer_bytes
from chipsec.hal.interrupts import Interrupts
from chipsec.library.exceptions import BadSMIDetected
from chipsec.helper.oshelper import OsHelper

TAGS = [SMM]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'TOOLS', 'SMM', 'SMM_PTR']


#################################################################
# Fuzzing configuration
#################################################################

#
# Logging option
#

# False - better performance, True - better results tracking
DUMP_MEMORY_ON_DETECT = False
# False - better performance, True - better results tracking
FLUSH_OUTPUT_ALWAYS = False
# makes sure SMI code is logged in case of a crash
FLUSH_OUTPUT_BEFORE_SMI = True
FLUSH_OUTPUT_AFTER_SMI = True
# dump all registers in log before every SMI (True - large size of log file)
DUMP_GPRS_EVERY_SMI = True

#
# SMI fuzzing options
#

# stop fuzzing after the first potential issue detected
FUZZ_BAIL_ON_1ST_DETECT = True

# Consider SMI handler subfunctions are passed in RCX GP register
# Fuzz RCX as SMI subfunctions: from 0 to MAX_SMI_FUNCTIONS
# False - better performance, True - smarter fuzzing
FUZZ_SMI_FUNCTIONS_IN_ECX = True
MAX_SMI_FUNCTIONS = 0x10

# Max value of the value written to SMI data port (0xB3)
MAX_SMI_DATA = 0x100

# Pass the pointer to SMI handlers in all general-purpose registers
# rather than in one register
# True - faster, False - gives you specific GPR that the vulnerable SMI handler is consuming
#
PTR_IN_ALL_GPRS = False

# SMI handler may take a pointer/PA from (some offset of off) address passed in GPRs and write to it
# Treat contents at physical address passed in GPRs as pointers and check contents at that pointer
# If they changed, SMI handler might have modified them
# MODE_SECOND_ORDER_BUFFER  = True

# Max offset of the pointer (physical address)
# of the 2nd order buffer written in the memory buffer passed to SMI
MAX_PTR_OFFSET_IN_BUFFER = 0x20

# very obscure option, don't even try to understand
GPR_2ADDR = False

# Defines the threshold in standard deviations at which the SMI call is
# considered long-running
OUTLIER_STD_DEV = 2

# Number of samples used for initial calibration
SCAN_CALIB_SAMPLES = 50

# SMI count MSR
MSR_SMI_COUNT = 0x00000034

#
# Defaults
#
_FILL_VALUE_QWORD = 0x5A5A5A5A5A5A5A5A
_FILL_VALUE_BYTE = 0x5A
_SMI_CODE_DATA = 0x0
_MEM_FILL_VALUE = b'\x11'
_MEM_FILL_SIZE = 0x500
_MAX_ALLOC_PA = 0xFFFFFFFF
_DEFAULT_GPRS = {'rax': _FILL_VALUE_QWORD, 'rbx': _FILL_VALUE_QWORD, 'rcx': _FILL_VALUE_QWORD, 'rdx': _FILL_VALUE_QWORD, 'rsi': _FILL_VALUE_QWORD, 'rdi': _FILL_VALUE_QWORD}

_pth = 'smm_ptr'


class gprs_info:
    def __init__(self, value):
        self.gprs = value

    def __repr__(self):
        return f"rax={self.gprs['rax']:02X} rbx={self.gprs['rbx']:02X} rcx={self.gprs['rcx']:02X}, " \
               f"rdx={self.gprs['rdx']:02X} rsi={self.gprs['rsi']:02X} rdi={self.gprs['rdi']:02X}"


class smi_info:
    def __init__(self, value):
        self.duration = value
        self.gprs = _DEFAULT_GPRS
        self.code = None
        self.data = None

    def update(self, value, code, data, gprs):
        self.duration = value
        self.gprs = gprs
        self.code = code
        self.data = data

    def get_info(self):
        if self.code is None:
            return None
        else:
            return f'duration {self.duration} code {self.code:02X} data {self.data:02X} ({gprs_info(self.gprs)})'


class smi_stats:
    def __init__(self):
        self.clear()

    def clear(self):
        self.count = 0
        self.mean = 0
        self.m2 = 0
        self.stdev = 0
        self.outliers = 0

    #
    # Computes the standard deviation using the Welford's online algorithm
    #
    def update_stats(self, duration):
        self.count += 1
        difference = duration - self.mean
        self.mean += difference / self.count
        self.m2 += difference * (duration - self.mean)
        variance = self.m2 / self.count
        self.stdev = math.sqrt(variance)

    def get_info(self):
        info = f'average {round(self.mean)} stddev {self.stdev:.2f} checked {self.count}'
        return info

    #
    # Combines the statistics of the two data sets using parallel variance computation
    #
    def combine(self, partial):
        self.outliers += partial.outliers
        total_count = self.count + partial.count
        difference = partial.mean - self.mean
        self.mean = (self.mean * self.count + partial.mean * partial.count) / total_count
        self.m2 += partial.m2 + difference**2 * self.count * partial.count / total_count
        self.count = total_count
        variance = self.m2 / self.count
        self.stdev = math.sqrt(variance)


class scan_track:
    def __init__(self):
        self.current_smi_stats = smi_stats()
        self.combined_smi_stats = smi_stats()
        self.clear()
        self.helper = OsHelper().get_default_helper()
        self.helper.init()
        self.smi_count = self.get_smi_count()

    def __del__(self):
        self.helper.close()

    def get_smi_count(self):
        count = -1
        #
        # The SMI count is the same on all CPUs
        #
        cpu = 0
        try:
            count = self.helper.read_msr(cpu, MSR_SMI_COUNT)
            count = count[1] << 32 | count[0]
        except UnimplementedAPIError:
            pass
        return count

    def valid_smi_count(self):
        valid = False
        count = self.get_smi_count()
        if (count == -1):
            return True
        elif (count == self.smi_count + 1):
            valid = True
        self.smi_count = count
        if not valid:
            sys.stderr.write('SMI contention detected')
        return valid

    def find_address_in_regs(self, gprs):
        for key, value in gprs.items():
            if key == 'rcx':
                continue
            if value != _FILL_VALUE_QWORD:
                return key

    def clear(self):
        self.outlier = smi_info(0)
        self.code = None
        self.contents_changed = False
        self.needs_calibration = True
        self.calib_samples = 0
        self.current_smi_stats.clear()

    def add(self, duration, code, data, gprs, contents_changed=False):
        if not self.code:
            self.code = code
        outlier = self.is_outlier(duration)
        if not outlier:
            self.current_smi_stats.update_stats(duration)
        elif self.is_slow_outlier(duration):
            self.current_smi_stats.outliers += 1
            self.outlier.update(duration, code, data, gprs.copy())
        self.contents_changed = contents_changed

    def update_calibration(self, duration):
        if not self.needs_calibration:
            return
        self.current_smi_stats.update_stats(duration)
        self.calib_samples += 1
        if self.calib_samples >= SCAN_CALIB_SAMPLES:
            self.needs_calibration = False

    def is_slow_outlier(self, value):
        ret = False
        if value > self.current_smi_stats.mean + OUTLIER_STD_DEV * self.current_smi_stats.stdev:
            ret = True
        if self.combined_smi_stats.count and \
            value > self.combined_smi_stats.mean + OUTLIER_STD_DEV * self.combined_smi_stats.stdev:
            ret = True
        return ret

    def is_fast_outlier(self, value):
        ret = False
        if value < self.current_smi_stats.mean - OUTLIER_STD_DEV * self.current_smi_stats.stdev:
            ret = True
        if self.combined_smi_stats.count and \
            value < self.combined_smi_stats.mean - OUTLIER_STD_DEV * self.combined_smi_stats.stdev:
            ret = True
        return ret

    def is_outlier(self, value):
        if self.needs_calibration:
            return False
        ret = False
        if self.is_slow_outlier(value):
            ret = True
        if self.is_fast_outlier(value):
            ret = True
        return ret

    def skip(self):
        return self.current_smi_stats.outliers or self.contents_changed

    def found_outlier(self):
        return bool(self.current_smi_stats.outliers)

    def get_total_outliers(self):
        return self.combined_smi_stats.outliers

    def get_info(self):
        info = self.current_smi_stats.get_info()
        if self.current_smi_stats.outliers:
            info += f'\n    Identified outlier: {self.outlier.get_info()}'
        return info

    def log_smi_result(self, logger):
        msg = f'SCANNED: SMI# {self.code:02X} {self.get_info()}'
        if self.found_outlier():
            logger.log_important(msg)
        else:
            logger.log(f'[*] {msg}')

    def update_combined_stats(self):
        self.combined_smi_stats.combine(self.current_smi_stats)


class smi_desc:
    def __init__(self):
        self.smi_code = None
        self.smi_data = None
        self.name = 'smi'
        self.desc = ''
        self.gprs = _DEFAULT_GPRS
        self.ptr_in_buffer = False
        self.ptr = None
        self.ptr_offset = 0
        self.sig = None
        self.sig_offset = 0


def DIFF(s, t, sz):
    return [pos for pos in range(sz) if s[pos] != t[pos]]


def FILL_BUFFER(_fill_byte, _fill_size, _ptr_in_buffer, _ptr, _ptr_offset, _sig, _sig_offset):
    fill_buf = _fill_byte * _fill_size
    if _ptr_in_buffer and (_ptr is not None):
        fill_buf = fill_buf[:_ptr_offset] + struct.pack('=I', _ptr & 0xFFFFFFFF) + fill_buf[_ptr_offset + 4:]
    if _sig is not None:
        fill_buf = fill_buf[:_sig_offset] + _sig + fill_buf[_sig_offset + len(_sig):]
    return fill_buf


class smm_ptr(BaseModule):
    def __init__(self):
        BaseModule.__init__(self)
        self.interrupts = Interrupts(self.cs)
        self.is_check_memory = True
        self.test_ptr_in_buffer = False
        self.fill_byte = _MEM_FILL_VALUE
        self.fill_size = _MEM_FILL_SIZE

    def is_supported(self):
        return True

    def fill_memory(self, _addr, is_ptr_in_buffer, _ptr, _ptr_offset, _sig, _sig_offset):
        #
        # Fill in contents at PA = _addr with known pattern to check later if any SMI handler modifies them
        #
        fill_buf = FILL_BUFFER(self.fill_byte, self.fill_size, is_ptr_in_buffer, _ptr, _ptr_offset, _sig, _sig_offset)

        s = f'[*] Writing 0x{self.fill_size:X} bytes at 0x{_addr:016X}'
        if is_ptr_in_buffer:
            s += f' -> PTR at +0x{_ptr_offset:X}'
        if _sig is not None:
            s += f' -> SIG at +0x{_sig_offset:X}'
        self.logger.log(s)
        self.cs.mem.write_physical_mem(_addr, self.fill_size, fill_buf)

        if self.logger.VERBOSE:
            self.logger.log(f'Filling in contents at PA 0x{_addr:016X}:')
            print_buffer_bytes(fill_buf, 16)

        if is_ptr_in_buffer and _ptr is not None:
            self.logger.log(f'[*] Writing buffer at PA 0x{_ptr:016X} with 0x{self.fill_size:X} bytes \'{self.fill_byte}\'')
            self.cs.mem.write_physical_mem(_ptr, self.fill_size, self.fill_byte * self.fill_size)

        return True

    def send_smi(self, thread_id, smi_code, smi_data, name, desc, rax, rbx, rcx, rdx, rsi, rdi):
        self.logger.log(f'    > SMI {smi_code:02X} (data: {smi_data:02X})')
        if DUMP_GPRS_EVERY_SMI:
            self.logger.log(f'      RAX: 0x{rax:016X}')
            self.logger.log(f'      RBX: 0x{rbx:016X}')
            self.logger.log(f'      RCX: 0x{rcx:016X}')
            self.logger.log(f'      RDX: 0x{rdx:016X}')
            self.logger.log(f'      RSI: 0x{rsi:016X}')
            self.logger.log(f'      RDI: 0x{rdi:016X}')
        self.interrupts.send_SW_SMI(thread_id, smi_code, smi_data, rax, rbx, rcx, rdx, rsi, rdi)
        return True

    def send_smi_timed(self, thread_id, smi_code, smi_data, name, desc, rax, rbx, rcx, rdx, rsi, rdi):
        self.logger.log(f'    > SMI {smi_code:02X} (data: {smi_data:02X})')
        if DUMP_GPRS_EVERY_SMI:
            self.logger.log(f'      RAX: 0x{rax:016X}')
            self.logger.log(f'      RBX: 0x{rbx:016X}')
            self.logger.log(f'      RCX: 0x{rcx:016X}')
            self.logger.log(f'      RDX: 0x{rdx:016X}')
            self.logger.log(f'      RSI: 0x{rsi:016X}')
            self.logger.log(f'      RDI: 0x{rdi:016X}')
        ret = self.interrupts.send_SW_SMI_timed(thread_id, smi_code, smi_data, rax, rbx, rcx, rdx, rsi, rdi)
        duration = ret[7]
        return (True, duration)

    def check_memory(self, _addr, _smi_desc, fn, restore_contents=False):
        _ptr = _smi_desc.ptr
        filler = self.fill_byte * self.fill_size
        #
        # Check if contents have changed at physical address passed in GPRs to SMI handler
        # If changed, SMI handler might have written to that address
        #
        self.logger.log('    < Checking buffers')

        expected_buf = FILL_BUFFER(self.fill_byte, self.fill_size, _smi_desc.ptr_in_buffer, _smi_desc.ptr, _smi_desc.ptr_offset, _smi_desc.sig, _smi_desc.sig_offset)
        buf = self.cs.mem.read_physical_mem(_addr, self.fill_size)
        differences = DIFF(expected_buf, buf, self.fill_size)
        _changed = len(differences) > 0

        if self.logger.VERBOSE:
            self.logger.log(f'Checking contents at PA 0x{_addr:016X}:')
            print_buffer_bytes(buf, 16)
            self.logger.log('Expected contents:')
            print_buffer_bytes(expected_buf, 16)

        if _changed:
            self.logger.log(f'    Contents changed at 0x{_addr:016X} +{differences}')
            if restore_contents:
                self.logger.log(f'    Restoring 0x{self.fill_size:X} bytes at 0x{_addr:016X}')
                self.cs.mem.write_physical_mem(_addr, self.fill_size, expected_buf)
            if DUMP_MEMORY_ON_DETECT:
                _pth_smi = os.path.join(_pth, f'{_smi_desc.smi_code:X}_{_smi_desc.name}')
                if not os.path.exists(_pth_smi):
                    os.makedirs(_pth_smi)
                _f = os.path.join(_pth_smi, fn + '.dmp')
                self.logger.log(f'    Dumping buffer to \'{_f}\'')
                write_file(_f, buf)

        _changed1 = False
        expected_buf = filler
        if _smi_desc.ptr_in_buffer and (_ptr is not None):
            buf1 = self.cs.mem.read_physical_mem(_ptr, self.fill_size)
            differences1 = DIFF(expected_buf, buf1, self.fill_size)
            _changed1 = len(differences1) > 0

            if self.logger.VERBOSE:
                self.logger.log(f'Checking contents at PA 0x{_ptr:016X}:')
                print_buffer_bytes(buf1, 16)

            if _changed1:
                self.logger.log(f'    Contents changed at 0x{_ptr:016X} +{differences1}')
                if restore_contents:
                    self.logger.log(f'    Restoring 0x{self.fill_size:X} bytes at PA 0x{_ptr:016X}')
                    self.cs.mem.write_physical_mem(_ptr, self.fill_size, expected_buf)
                if DUMP_MEMORY_ON_DETECT:
                    _pth_smi = os.path.join(_pth, f'{_smi_desc.smi_code:X}_{_smi_desc.name}')
                    if not os.path.exists(_pth_smi):
                        os.makedirs(_pth_smi)
                    _f = os.path.join(_pth_smi, fn + (f'_ptr{_smi_desc.ptr_offset:X}.dmp'))
                    self.logger.log(f'    Dumping buffer to \'{_f}\'')
                    write_file(_f, buf1)

        return (_changed or _changed1)

    def smi_fuzz_iter(self, thread_id, _addr, _smi_desc, fill_contents=True, restore_contents=False, scan=None):
        #
        # Fill memory buffer if not in 'No Fill' mode
        #
        if self.is_check_memory and fill_contents:
            self.fill_memory(_addr, _smi_desc.ptr_in_buffer, _smi_desc.ptr, _smi_desc.ptr_offset, _smi_desc.sig, _smi_desc.sig_offset)

        if FLUSH_OUTPUT_BEFORE_SMI:
            self.logger.flush()
        #
        # Invoke SW SMI Handler
        #
        _rax = _smi_desc.gprs['rax']
        _rbx = _smi_desc.gprs['rbx']
        _rcx = _smi_desc.gprs['rcx']
        _rdx = _smi_desc.gprs['rdx']
        _rsi = _smi_desc.gprs['rsi']
        _rdi = _smi_desc.gprs['rdi']
        if not scan:
            self.send_smi(thread_id, _smi_desc.smi_code, _smi_desc.smi_data, _smi_desc.name, _smi_desc.desc, _rax, _rbx, _rcx, _rdx, _rsi, _rdi)
        else:
            while True:
                _, duration = self.send_smi_timed(thread_id, _smi_desc.smi_code, _smi_desc.smi_data, _smi_desc.name, _smi_desc.desc, _rax, _rbx, _rcx, _rdx, _rsi, _rdi)
                if not scan.valid_smi_count():
                    continue
                if scan.needs_calibration:
                    scan.update_calibration(duration)
                    continue
                else:
                    break
            #
            # Re-do the call if it was identified as an outlier, due to periodic SMI delays
            #
            if scan.is_outlier(duration):
                while True:
                    _, duration = self.send_smi_timed(thread_id, _smi_desc.smi_code, _smi_desc.smi_data, _smi_desc.name, _smi_desc.desc, _rax, _rbx, _rcx, _rdx, _rsi, _rdi)
                    if scan.valid_smi_count():
                        break
        #
        # Check memory buffer if not in 'No Fill' mode
        #
        contents_changed = False
        if self.is_check_memory:
            fn = f'{_smi_desc.smi_data:X}-a{_rax:X}_b{_rbx:X}_c{_rcx:X}_d{_rdx:X}_si{_rsi:X}_di{_rdi:X}'
            contents_changed = self.check_memory(_addr, _smi_desc, fn, restore_contents)
            if contents_changed:
                msg = f'DETECTED: SMI# {_smi_desc.smi_code:X} data {_smi_desc.smi_data:X} (rax={_rax:X} rbx={_rbx:X} rcx={_rcx:X} rdx={_rdx:X} rsi={_rsi:X} rdi={_rdi:X})'
                self.logger.log_important(msg)
                if FUZZ_BAIL_ON_1ST_DETECT and not scan:
                    raise BadSMIDetected(msg)

        if scan:
            scan.add(duration, _smi_desc.smi_code, _smi_desc.smi_data, _smi_desc.gprs, contents_changed)

        if FLUSH_OUTPUT_AFTER_SMI:
            self.logger.flush()

        return contents_changed

    def test_config(self, thread_id, _smi_config_fname, _addr, _addr1):
        #
        # Parse SMM config file describing SMI handlers and their call arguments
        # Then invoke SMI handlers
        #
        fcfg = open(_smi_config_fname, 'r')
        self.logger.log(f'\n[*] >>> Testing SMI handlers defined in \'{_smi_config_fname}\'..')

        bad_ptr_cnt = 0
        _smi_desc = smi_desc()
        for line in fcfg:
            if '' == line.strip():
                self.logger.log(f'\n[*] Testing SMI# 0x{_smi_desc.smi_code:02X} (data: 0x{_smi_desc.smi_data:02X}) {_smi_desc.name} ({_smi_desc.desc})')
                if self.smi_fuzz_iter(thread_id, _addr, _smi_desc):
                    bad_ptr_cnt += 1
                _smi_desc = None
                _smi_desc = smi_desc()
            else:
                name, var = line.strip().partition('=')[::2]
                _n = name.strip().lower()
                if 'name' == _n:
                    _smi_desc.name = var
                elif 'desc' == _n:
                    _smi_desc.desc = var
                elif 'smi_code' == _n:
                    _smi_desc.smi_code = int(var, 16) if '*' != var else _SMI_CODE_DATA
                elif 'smi_data' == _n:
                    _smi_desc.smi_data = int(var, 16) if '*' != var else _SMI_CODE_DATA
                elif 'ptr_offset' == _n:
                    _smi_desc.ptr_in_buffer = True
                    _smi_desc.ptr_offset = int(var, 16)
                    _smi_desc.ptr = _addr1
                elif 'sig' == _n:
                    _smi_desc.sig = bytearray.fromhex(var)
                elif 'sig_offset' == _n:
                    _smi_desc.sig_offset = int(var, 16)
                else:
                    _smi_desc.gprs[_n] = (_addr if 'PTR' == var else (_FILL_VALUE_BYTE if 'VAL' == var else int(var, 16))) if '*' != var else _FILL_VALUE_QWORD

        return bad_ptr_cnt

    def test_fuzz(self, thread_id, smic_start, smic_end, _addr, _addr1, scan_mode=False):

        scan = None
        if scan_mode:
            scan = scan_track()

        gpr_value = ((_addr << 32) | _addr) if GPR_2ADDR else _addr

        gprs_addr = {'rax': gpr_value, 'rbx': gpr_value, 'rcx': gpr_value, 'rdx': gpr_value, 'rsi': gpr_value, 'rdi': gpr_value}
        gprs_fill = {'rax': _FILL_VALUE_QWORD, 'rbx': _FILL_VALUE_QWORD, 'rcx': _FILL_VALUE_QWORD, 'rdx': _FILL_VALUE_QWORD, 'rsi': _FILL_VALUE_QWORD, 'rdi': _FILL_VALUE_QWORD}
        self.logger.log('\n[*] >>> Fuzzing SMI handlers..')
        self.logger.log('[*] AX in RAX will be overridden with values of SW SMI ports 0xB2/0xB3')
        self.logger.log('    DX in RDX will be overridden with value 0x00B2')

        bad_ptr_cnt = 0
        _smi_desc = smi_desc()
        _smi_desc.gprs = gprs_addr if PTR_IN_ALL_GPRS or scan_mode else gprs_fill
        self.logger.log(f'\n[*] Setting values of general purpose registers to 0x{_smi_desc.gprs["rax"]:016X}')
        max_ptr_off = 1

        if self.is_check_memory and self.test_ptr_in_buffer:
            _smi_desc.ptr_in_buffer = True
            _smi_desc.ptr = _addr1
            max_ptr_off = MAX_PTR_OFFSET_IN_BUFFER + 1

        # if we are not in fuzzmore mode, i.e. we are not testing the pointer within memory buffer
        # then this outer loop will only have 1 iteration
        for off in range(max_ptr_off):
            _smi_desc.ptr_offset = off
            self.logger.log(f'\n[*] Reloading buffer with PTR at offset 0x{off:X}..')
            if self.is_check_memory:
                self.fill_memory(_addr, _smi_desc.ptr_in_buffer, _smi_desc.ptr, _smi_desc.ptr_offset, None, None)

            for smi_code in range(smic_start, smic_end + 1, 1):
                _smi_desc.smi_code = smi_code
                for smi_data in range(MAX_SMI_DATA):
                    _smi_desc.smi_data = smi_data
                    self.logger.log(f'\n[*] Fuzzing SMI# 0x{smi_code:02X} (data: 0x{smi_data:02X})')
                    if FUZZ_SMI_FUNCTIONS_IN_ECX:
                        for _rcx in range(MAX_SMI_FUNCTIONS):
                            self.logger.log(f' >> Function (RCX): 0x{_rcx:016X}')
                            _smi_desc.gprs['rcx'] = _rcx
                            if PTR_IN_ALL_GPRS or scan_mode:
                                if self.smi_fuzz_iter(thread_id, _addr, _smi_desc, False, True, scan):
                                    bad_ptr_cnt += 1
                                if scan and scan.skip():
                                    break
                            else:
                                self.logger.log(f'    RBX: 0x{_addr:016X}')
                                _smi_desc.gprs['rbx'] = gpr_value
                                if self.smi_fuzz_iter(thread_id, _addr, _smi_desc, False, True):
                                    bad_ptr_cnt += 1
                                _smi_desc.gprs['rbx'] = _FILL_VALUE_QWORD

                                self.logger.log(f'    RSI: 0x{_addr:016X}')
                                _smi_desc.gprs['rsi'] = gpr_value
                                if self.smi_fuzz_iter(thread_id, _addr, _smi_desc, False, True):
                                    bad_ptr_cnt += 1
                                _smi_desc.gprs['rsi'] = _FILL_VALUE_QWORD

                                self.logger.log(f'    RDI: 0x{_addr:016X}')
                                _smi_desc.gprs['rdi'] = gpr_value
                                if self.smi_fuzz_iter(thread_id, _addr, _smi_desc, False, True):
                                    bad_ptr_cnt += 1
                                _smi_desc.gprs['rdi'] = _FILL_VALUE_QWORD
                    else:
                        if PTR_IN_ALL_GPRS or scan_mode:
                            if self.smi_fuzz_iter(thread_id, _addr, _smi_desc, False, True, scan):
                                bad_ptr_cnt += 1
                        else:
                            self.logger.log(f'    RBX: 0x{_addr:016X}')
                            _smi_desc.gprs['rbx'] = gpr_value
                            if self.smi_fuzz_iter(thread_id, _addr, _smi_desc, False, True):
                                bad_ptr_cnt += 1
                            _smi_desc.gprs['rbx'] = _FILL_VALUE_QWORD

                            self.logger.log(f'    RCX: 0x{_addr:016X}')
                            _smi_desc.gprs['rcx'] = gpr_value
                            if self.smi_fuzz_iter(thread_id, _addr, _smi_desc, False, True):
                                bad_ptr_cnt += 1
                            _smi_desc.gprs['rcx'] = _FILL_VALUE_QWORD

                            self.logger.log(f'    RSI: 0x{_addr:016X}')
                            _smi_desc.gprs['rsi'] = gpr_value
                            if self.smi_fuzz_iter(thread_id, _addr, _smi_desc, False, True):
                                bad_ptr_cnt += 1
                            _smi_desc.gprs['rsi'] = _FILL_VALUE_QWORD

                            self.logger.log(f'    RDI: 0x{_addr:016X}')
                            _smi_desc.gprs['rdi'] = gpr_value
                            if self.smi_fuzz_iter(thread_id, _addr, _smi_desc, False, True):
                                bad_ptr_cnt += 1
                            _smi_desc.gprs['rdi'] = _FILL_VALUE_QWORD
                    if scan and scan.skip():
                        break
                if scan_mode:
                    scan.log_smi_result(self.logger)
                    scan.update_combined_stats()
                    scan.clear()

        return bad_ptr_cnt, scan

    def run(self, module_argv):
        self.logger.start_test('A tool to test SMI handlers for pointer validation vulnerabilities')
        self.logger.log('Usage: chipsec_main -m tools.smm.smm_ptr [ -a <mode>,<config_file>|<smic_start:smic_end>,<size>,<address> ]')
        self.logger.log('  mode          SMI handlers testing mode')
        self.logger.log('    = config    use SMI configuration file <config_file>')
        self.logger.log('    = fuzz      fuzz all SMI handlers with code in the range <smic_start:smic_end>')
        self.logger.log('    = fuzzmore  fuzz mode + pass `2nd-order` pointers within buffer to SMI handlers')
        self.logger.log('    = scan      fuzz mode + time measurement to identify SMIs that trigger long-running code paths')
        self.logger.log('  size          size of the memory buffer (in Hex)')
        self.logger.log('  address       physical address of memory buffer to pass in GP regs to SMI handlers (in Hex)')
        self.logger.log('    = smram     pass address of SMRAM base (system may hang in this mode!)\n')

        test_mode = 'config'
        _smi_config_fname = 'chipsec/modules/tools/smm/smm_config.ini'
        _addr = None
        _addr1 = None
        thread_id = 0x0

        global DUMP_GPRS_EVERY_SMI
        smic_start = 0
        smic_end = 0
        if len(module_argv) > 1:
            test_mode = module_argv[0].lower()
            if test_mode == 'config':
                _smi_config_fname = module_argv[1]
            elif test_mode in ['fuzz', 'fuzzmore', 'scan']:
                smic_arr = module_argv[1].split(':')
                smic_start = int(smic_arr[0], 16)
                smic_end = int(smic_arr[1], 16)
                if test_mode == 'fuzzmore':
                    self.test_ptr_in_buffer = True
                    DUMP_GPRS_EVERY_SMI = False
                elif test_mode == 'scan':
                    self.test_ptr_in_buffer = False
            else:
                self.logger.log_error(f'Unknown fuzzing mode \'{module_argv[0]}\'')
                self.result.setStatusBit(self.result.status.UNSUPPORTED_OPTION)
                return self.result.getReturnCode(ModuleResult.ERROR)

        if len(module_argv) > 2:
            self.fill_size = int(module_argv[2], 16)
        if len(module_argv) > 3:
            if 'smram' == module_argv[3]:
                (_addr, _, _) = self.cs.cpu.get_SMRAM()
                self.is_check_memory = False
                self.logger.log(f'[*] Using SMRAM base address (0x{_addr:016X}) to pass to SMI handlers')
            else:
                _addr = int(module_argv[3], 16)
                self.logger.log(f'[*] Using address from command-line (0x{_addr:016X}) to pass to SMI handlers')
        else:
            (_, _addr) = self.cs.mem.alloc_physical_mem(self.fill_size, _MAX_ALLOC_PA)
            self.logger.log(f'[*] Allocated memory buffer (to pass to SMI handlers)       : 0x{_addr:016X}')

        if self.is_check_memory:
            (_, _addr1) = self.cs.mem.alloc_physical_mem(self.fill_size, _MAX_ALLOC_PA)
            self.logger.log(f'[*] Allocated 2nd buffer (address will be in the 1st buffer): 0x{_addr1:016X}')

        # @TODO: Need to check that SW/APMC SMI is enabled

        self.logger.log('\n[*] Configuration:')
        self.logger.log(f'    SMI testing mode          : {test_mode}')
        if test_mode == 'config':
            self.logger.log(f'    Config file           : {_smi_config_fname}')
        else:
            self.logger.log(f'    Range of SMI codes (B2)   : 0x{smic_start:02X}:0x{smic_end:02X}')
        self.logger.log(f'    Memory buffer pointer     : 0x{_addr:016X} (address passed in GP regs to SMI)')
        self.logger.log(f'    Filling/checking memory?  : {"YES" if self.is_check_memory else "NO"}')
        if self.is_check_memory:
            self.logger.log(f'      Second buffer pointer   : 0x{_addr1:016X} (address written to memory buffer)')
            self.logger.log(f'      Number of bytes to fill : 0x{self.fill_size:X}')
            self.logger.log(f'      Byte to fill with       : 0x{ord(self.fill_byte):X}')
        self.logger.log('    Additional options (can be changed in the source code):')
        self.logger.log(f'      Fuzzing SMI functions in ECX?          : {FUZZ_SMI_FUNCTIONS_IN_ECX:d}')
        self.logger.log(f'      Max value of SMI function in ECX       : 0x{MAX_SMI_FUNCTIONS:X}')
        self.logger.log(f'      Max value of SMI data (B3)             : 0x{MAX_SMI_DATA:X}')
        self.logger.log(f'      Max offset of the pointer in the buffer: 0x{MAX_PTR_OFFSET_IN_BUFFER:X}')
        self.logger.log(f'      Passing pointer in all GP registers?   : {PTR_IN_ALL_GPRS:d}')
        self.logger.log(f'      Default values of the registers        : 0x{_FILL_VALUE_QWORD:016X}')
        self.logger.log(f'      Dump all register values every SMI     : {DUMP_GPRS_EVERY_SMI:d}')
        self.logger.log(f'      Bail on first detection                : {FUZZ_BAIL_ON_1ST_DETECT:d}')

        self.logger.set_always_flush(FLUSH_OUTPUT_ALWAYS)
        if DUMP_MEMORY_ON_DETECT and not os.path.exists(_pth):
            os.makedirs(_pth)

        bad_ptr_cnt = 0
        scan_mode = False
        try:
            if 'config' == test_mode:
                bad_ptr_cnt = self.test_config(thread_id, _smi_config_fname, _addr, _addr1)
            elif test_mode in ['fuzz', 'fuzzmore']:
                bad_ptr_cnt, _ = self.test_fuzz(thread_id, smic_start, smic_end, _addr, _addr1)
            elif test_mode in ['scan']:
                scan_mode = True
                scan = None
                bad_ptr_cnt, scan = self.test_fuzz(thread_id, smic_start, smic_end, _addr, _addr1, True)
        except BadSMIDetected:
            bad_ptr_cnt = 1
            self.logger.log_important('Potentially bad SMI detected! Stopped fuzing (see FUZZ_BAIL_ON_1ST_DETECT option)')

        if scan_mode and scan:
            self.logger.log_good(f'<<< Done: found {scan.get_total_outliers()} long-running SMIs')
        if bad_ptr_cnt > 0:
            self.logger.log_bad(f'<<< Done: found {bad_ptr_cnt:d} potential occurrences of unchecked input pointers')
            self.result.setStatusBit(self.result.status.POTENTIALLY_VULNERABLE)
            self.res = self.result.getReturnCode(ModuleResult.FAILED)
        else:
            self.logger.log_good('<<< Done: did not find unchecked input pointers in tested SMI handlers')
            self.result.setStatusBit(self.result.status.SUCCESS)
            self.res = self.result.getReturnCode(ModuleResult.PASSED)

        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
x64/IA-64 Paging functionality including x86 page tables, Extended Page Tables (EPT) and VT-d page tables
"""

import sys
import struct
from typing import Dict, List, Optional, Any
import chipsec.library.defines as defines
from chipsec.library.logger import logger
from chipsec.library.exceptions import InvalidMemoryAddress

ADDR_MASK = defines.MASK_64b
MAXPHYADDR = 0x000FFFFFFFFFF000

SIZE_4KB = defines.BOUNDARY_4KB
SIZE_2MB = defines.BOUNDARY_2MB
SIZE_1GB = defines.BOUNDARY_1GB
ADDR_4KB = 0xFFFFFFFFFFFFF000 & MAXPHYADDR
ADDR_2MB = 0xFFFFFFFFFFE00000 & MAXPHYADDR
ADDR_1GB = 0xFFFFFFFFC0000000 & MAXPHYADDR

TranslationType = Dict[int, Dict[str, Any]]  # TODO: TypedDict (PEP589)


class c_translation:

    def __init__(self):
        self.translation: TranslationType = {}

    def is_translation_exist(self, addr: int, mask: int, size: str) -> bool:
        return ((addr & mask) in self.translation) and (self.translation[addr & mask]['size'] == size)

    def get_translation(self, addr: int) -> Optional[int]:
        if len(self.translation) == 0:
            return addr
        ADDR_4KB = 0xFFFFFFFFFFFFF000
        ADDR_2MB = 0xFFFFFFFFFFE00000
        ADDR_1GB = 0xFFFFFFFFC0000000
        if self.is_translation_exist(addr, ADDR_4KB, '4KB'):
            result = self.translation[addr & ADDR_4KB]['addr'] | (addr & ~ADDR_4KB)
        elif self.is_translation_exist(addr, ADDR_2MB, '2MB'):
            result = self.translation[addr & ADDR_2MB]['addr'] | (addr & ~ADDR_2MB)
        elif self.is_translation_exist(addr, ADDR_1GB, '1GB'):
            result = self.translation[addr & ADDR_1GB]['addr'] | (addr & ~ADDR_1GB)
        else:
            result = None
        return result

    def get_pages_by_physaddr(self, addr: int) -> List[Dict[str, int]]:
        SIZE = {'4KB': ADDR_4KB, '2MB': ADDR_2MB, '1GB': ADDR_1GB}
        result = []
        for i in self.translation.keys():
            page = self.translation[i]
            size = SIZE[page['size']]
            if (page['addr'] & size) == (addr & size):
                result.append(page)
        return result

    def get_address_space(self) -> int:
        total = 0
        mem_range = self.get_mem_range()
        for i in mem_range:
            total += i[1] - i[0]
        return total

    def get_mem_range(self, noattr: bool = False) -> List[List[int]]:
        SIZE = {'4KB': SIZE_4KB, '2MB': SIZE_2MB, '1GB': SIZE_1GB}
        perm = {self.translation[a]['addr']: self.translation[a] for a in self.translation.keys()}
        mem_range = []
        for addr in sorted(perm.keys()):
            attr = perm[addr]['attr']
            size = SIZE[perm[addr]['size']]
            if noattr:
                attr = ''
            if (mem_range == []):
                mem_range += [[addr, addr + size, attr]]
            elif (mem_range[-1][1] == addr) and (mem_range[-1][2] == attr):
                mem_range[-1][1] += size
            else:
                mem_range += [[addr, addr + size, attr]]
        return mem_range

    def add_page(self, virt: int, phys: int, size: str, attr: str) -> None:
        if size not in ['4KB', '2MB', '4MB', '1GB']:
            raise Exception('Invalid size!')
        self.translation[virt] = {'addr': phys, 'size': size, 'attr': attr}
        return

    def del_page(self, addr: int) -> None:
        if addr in self.translation:
            del self.translation[addr]
        return

    def expand_pages(self, exp_size: str) -> None:
        SIZE = {'1GB': '2MB', '2MB': '4KB'}
        for virt in self.translation.keys():
            size = self.translation[virt]['size']
            attr = self.translation[virt]['attr']
            phys = self.translation[virt]['addr']
            pgsize = (1 << 12) if size == '2MB' else (1 << 20)
            if size == exp_size:
                for i in range(512):
                    self.add_page(virt + i * pgsize, phys + i * pgsize, SIZE[exp_size], attr)
        return


class c_reverse_translation:

    def __init__(self, translation: TranslationType):
        self.reverse_translation: Dict[int, List[Dict[str, Any]]] = {}
        for virt in translation.keys():
            phys = translation[virt]['addr']
            size = translation[virt]['size']
            attr = translation[virt]['attr']
            if phys not in self.reverse_translation:
                self.reverse_translation[phys] = []
            self.reverse_translation[phys].append({'addr': virt, 'size': size, 'attr': attr})

    def get_reverse_translation(self, addr: int) -> List[Dict[str, Any]]:
        ADDR_4KB = 0xFFFFFFFFFFFFF000
        addr &= ADDR_4KB
        return self.reverse_translation[addr] if addr in self.reverse_translation else []


class c_paging_memory_access:

    def __init__(self, cs):
        self.cs = cs

    def readmem(self, name: str, addr: int, size: int = 4096) -> bytes:
        return self.cs.mem.read_physical_mem(addr, size)


class c_paging_with_2nd_level_translation(c_paging_memory_access):

    def __init__(self, cs):
        c_paging_memory_access.__init__(self, cs)
        self.translation_level2 = c_translation()

    def readmem(self, name: str, addr: int, size: int = 4096) -> bytes:
        phys = self.translation_level2.get_translation(addr)
        if phys is None:
            logger().log_hal('[paging] get_translation(): phys is None. Returning 0.')
            return b''
        if phys != addr:
            name += f'_0x{phys:08X}'
        return super(c_paging_with_2nd_level_translation, self).readmem(name, phys, size)


class c_paging(c_paging_with_2nd_level_translation, c_translation):
    def __init__(self, cs):
        c_paging_with_2nd_level_translation.__init__(self, cs)
        c_translation.__init__(self)
        # variables
        self.did = 0
        self.out = sys.stdout
        self.name = ''
        self.pt = {}
        self.pointer = None
        self.failure = False
        self.canonical_msb = 47

    def get_canonical(self, va: int) -> int:
        canonical_mask = (ADDR_MASK << (self.canonical_msb + 1)) & ADDR_MASK
        canonical_va = (va | canonical_mask) if (va >> self.canonical_msb) & 0x1 else va
        return canonical_va

    def get_field(self, entry: int, desc: Dict[str, int]) -> int:
        return (entry >> desc['offset']) & desc['mask']

    def set_field(self, value: int, desc: Dict[str, int]) -> int:
        return (value & desc['mask']) << desc['offset']

    def read_entries(self, info: str, addr: int, size: int = 8) -> List[Any]:
        data = self.readmem(f'{self.name}_{info}_0x{addr:08X}', addr, 0x1000)
        entries = struct.unpack('<512Q', data)
        if size == 16:
            entries = [[entries[i], entries[i + 1]] for i in range(0, 512, 2)]

        same = True
        for i in range(len(entries)):
            same = same and (entries[0] == entries[i])
        if same:
            return [entries[0]]
        return entries

    def print_info(self, name: str) -> None:
        logger().log(f'\n  {name} physical address ranges:')
        mem_range = self.get_mem_range()
        for index in range(len(mem_range)):
            i = mem_range[index]
            logger().log(f'    0x{i[0]:013X} - 0x{i[1] - 1:013X} {(i[1] - i[0]) >> 12:8d}  {i[2]}')

        logger().log(f'\n  {name} pages:')
        for i in sorted(self.pt.keys()):
            logger().log(f'    0x{i:013X}  {self.pt[i]}')
        logger().log('\n')
        logger().log(f'  {name} size: {len(self.pt.keys()) * 4:d} KB, address space: {self.get_address_space() >> 20:d} MB')
        return

    def check_misconfig(self, addr_list: List[int]) -> None:
        addr_list = [x & MAXPHYADDR for x in addr_list]
        mem_range = self.get_mem_range()
        for addr in addr_list:
            for i in range(len(mem_range)):
                if (mem_range[i][0] <= addr) and (addr < mem_range[i][1]):
                    logger().log_hal(f'*** WARNING: PAGE TABLES MISCONFIGURATION  0x{addr:013X}')
        return

    def save_configuration(self, path: str) -> None:
        with open(path, 'w') as cfg:
            try:
                cfg.write(str(self.translation_level2.translation) + '\n')
                cfg.write(str(self.translation) + '\n')
                cfg.write(str(self.pt))
            except:
                logger().log_hal(f'[paging] Error saving: {path}')
        return

    def load_configuration(self, path: str) -> None:
        with open(path, 'r') as cfg:
            try:
                self.translation_level2.translation = eval(cfg.readline())
                self.translation = eval(cfg.readline())
                self.pt = eval(cfg.readline())
            except:
                logger().log_hal(f'[paging] Error loading: {path}')
        return

    def read_pt_and_show_status(self, path: str, name: str, ptr: int) -> None:
        logger().log_hal(f'[paging] Reading {name} page tables at 0x{ptr:016X}...')
        try:
            self.read_page_tables(ptr)
        except InvalidMemoryAddress:
            self.translation_level2.translation = {}
            self.translation = {}
            self.pt = {}
            self.failure = True
            if logger().HAL:
                logger().log_error(f'    Invalid {name} Page Tables!')
        else:
            self.print_info(f'[paging] {name} page tables')
            self.failure = False
            logger().log_hal(f'[paging] size: {len(self.pt.keys()) * 4:d} KB, address space: {self.get_address_space() >> 20:d} MB')
        return

    def read_page_tables(self, entry: int):
        raise Exception("Function needs to be implemented by child class")


class c_4level_page_tables(c_paging):

    def __init__(self, cs):
        c_paging.__init__(self, cs)
        # constants
        self.PHYSICAL_ADDR_NAME = ''
        self.PML4_INDX = {'mask': 0x1FF, 'offset': 39}
        self.PDPT_INDX = {'mask': 0x1FF, 'offset': 30}
        self.PD_INDX = {'mask': 0x1FF, 'offset': 21}
        self.PT_INDX = {'mask': 0x1FF, 'offset': 12}
        self.PT_NAME = ['EPTP', 'PML4E', 'PDPTE', 'PDE', 'PTE']
        self.PT_SIZE = ['', '', '1GB', '2MB', '4KB']

    def get_virt_addr(self, pml4e_index: int, pdpte_index: int = 0, pde_index: int = 0, pte_index: int = 0) -> int:
        ofs1 = self.set_field(pml4e_index, self.PML4_INDX)
        ofs2 = self.set_field(pdpte_index, self.PDPT_INDX)
        ofs3 = self.set_field(pde_index, self.PD_INDX)
        ofs4 = self.set_field(pte_index, self.PT_INDX)
        return (ofs1 | ofs2 | ofs3 | ofs4)

    def print_entry(self, lvl: int, pa: int, va: int = 0, perm: str = '') -> None:
        canonical_va = self.get_canonical(va)
        info = f'  {"  " * lvl}{self.PT_NAME[lvl]:6}: {pa:013X}'
        if perm != '':
            size = self.PT_SIZE[lvl]
            info += f' - {size} PAGE  {perm}'
            info = info.ljust(64)
            if pa == va:
                info += '1:1 mapping'
            else:
                info += f'{self.PHYSICAL_ADDR_NAME}: {canonical_va:013X}'

            self.add_page(canonical_va, pa, size, perm)

        logger().log(info)
        return

    def read_page_tables(self, ptr: int) -> None:
        addr = ptr & ADDR_4KB
        self.pointer = addr
        self.pt = {addr: 'pml4'}
        self.translation = {}
        self.print_entry(0, addr)
        self.read_pml4(addr)
        return

    def is_present(self, entry: int) -> int:
        return entry & defines.BIT0

    def is_bigpage(self, entry: int) -> int:
        return entry & defines.BIT7

    def read_pml4(self, addr: int) -> None:
        pml4 = self.read_entries('pml4', addr)
        for pml4e_index in range(len(pml4)):
            pml4e = pml4[pml4e_index]
            if self.is_present(pml4e):
                addr = pml4e & ADDR_4KB
                self.pt[addr] = 'pdpt'
                self.print_entry(1, addr)
                self.read_pdpt(addr, pml4e_index)
        return

    def get_attr(self, entry: int) -> str:
        ret = ''
        if entry & defines.BIT1:
            ret += 'W'
        else:
            ret += "R"
        if entry & defines.BIT2:
            ret += 'U'
        else:
            ret += 'S'
        return ret

    def read_pdpt(self, addr: int, pml4e_index: int) -> None:
        pdpt = self.read_entries('pdpt', addr)
        for pdpte_index in range(len(pdpt)):
            pdpte = pdpt[pdpte_index]
            if self.is_present(pdpte):
                if self.is_bigpage(pdpte):
                    virt = self.get_virt_addr(pml4e_index, pdpte_index)
                    phys = pdpte & ADDR_1GB
                    self.print_entry(2, phys, virt, self.get_attr(pdpte))
                else:
                    addr = pdpte & ADDR_4KB
                    self.pt[addr] = 'pd'
                    self.print_entry(2, addr)
                    self.read_pd(addr, pml4e_index, pdpte_index)
        return

    def read_pd(self, addr: int, pml4e_index: int, pdpte_index: int) -> None:
        pd = self.read_entries('pd', addr)
        for pde_index in range(len(pd)):
            pde = pd[pde_index]
            if self.is_present(pde):
                if self.is_bigpage(pde):
                    virt = self.get_virt_addr(pml4e_index, pdpte_index, pde_index)
                    phys = pde & ADDR_2MB
                    self.print_entry(3, phys, virt, self.get_attr(pde))
                else:
                    addr = pde & ADDR_4KB
                    self.pt[addr] = 'pt'
                    self.print_entry(3, addr)
                    self.read_pt(addr, pml4e_index, pdpte_index, pde_index)
        return

    def read_pt(self, addr: int, pml4e_index: int, pdpte_index: int, pde_index: int) -> None:
        pt = self.read_entries('pt', addr)
        for pte_index in range(len(pt)):
            pte = pt[pte_index]
            if self.is_present(pte):
                virt = self.get_virt_addr(pml4e_index, pdpte_index, pde_index, pte_index)
                phys = pte & ADDR_4KB
                self.print_entry(4, phys, virt, self.get_attr(pte))
        return

    def read_entry_by_virt_addr(self, virt: int) -> Dict[str, Any]:
        if self.pointer is None:
            raise Exception('Page Table pointer is undefined!')
        addr = self.pointer
        pml4 = self.read_entries('pml4', addr)
        pml4e = pml4[self.get_field(virt, self.PML4_INDX)]
        if self.is_present(pml4e):
            addr = pml4e & ADDR_4KB
            pdpt = self.read_entries('pdpt', addr)
            pdpte = pdpt[self.get_field(virt, self.PDPT_INDX)]
            if self.is_present(pdpte):
                if self.is_bigpage(pdpte):
                    addr = (pdpte & ADDR_1GB) | (virt & ~ADDR_1GB)
                    return {'addr': addr, 'attr': self.get_attr(pdpte), 'size': '1GB'}
                else:
                    addr = pdpte & ADDR_4KB
                    pd = self.read_entries('pd', addr)
                    pde = pd[self.get_field(virt, self.PD_INDX)]
                    if self.is_present(pde):
                        if self.is_bigpage(pde):
                            addr = (pde & ADDR_2MB) | (virt & ~ADDR_2MB)
                            return {'addr': addr, 'attr': self.get_attr(pde), 'size': '2MB'}
                        else:
                            addr = pde & ADDR_4KB
                            pt = self.read_entries('pt', addr)
                            pte = pt[self.get_field(virt, self.PT_INDX)]
                            if self.is_present(pte):
                                addr = (pte & ADDR_4KB) | (virt & ~ADDR_4KB)
                                return {'addr': addr, 'attr': self.get_attr(pte), 'size': '4KB'}
        return {'addr': 0, 'attr': '', 'size': ''}


class c_ia32e_page_tables(c_4level_page_tables):

    def __init__(self, cs):
        c_4level_page_tables.__init__(self, cs)
        # constants
        self.PHYSICAL_ADDR_NAME = 'VA'
        self.PT_NAME = ['CR3P', 'PML4E', 'PDPTE', 'PDE', 'PTE']
        self.P = {'mask': 0x1, 'offset': 0}
        self.RW = {'mask': 0x1, 'offset': 1}
        self.US = {'mask': 0x1, 'offset': 2}
        self.BIGPAGE = {'mask': 0x1, 'offset': 7}

    def is_present(self, entry: int) -> bool:
        return self.get_field(entry, self.P) != 0

    def is_bigpage(self, entry: int) -> bool:
        return self.get_field(entry, self.BIGPAGE) != 0

    def get_attr(self, entry: int) -> str:
        RW_DESC = ['R', 'W']
        US_DESC = ['S', 'U']
        return f'{RW_DESC[self.get_field(entry, self.RW)]} {US_DESC[self.get_field(entry, self.US)]}'


class c_pae_page_tables(c_ia32e_page_tables):

    def __init__(self, cs):
        c_ia32e_page_tables.__init__(self, cs)
        # constants
        self.PML4_INDX = {'mask': 0x000, 'offset': 39}
        self.PDPT_INDX = {'mask': 0x003, 'offset': 30}
        self.PT_NAME = ['', 'CR3', 'PDPTE', 'PDE', 'PTE']

    def read_page_tables(self, ptr: int) -> None:
        addr = ptr & ADDR_4KB
        self.pointer = addr
        self.pt = {addr: 'pdpt'}
        self.translation = {}
        self.print_entry(1, addr)
        self.read_pdpt(addr, None)
        return

    def read_pml4(self, addr: int):
        raise Exception('PAE Page tables have no PML4!')

    def read_pdpt(self, addr: int, pml4e_index: Optional[int] = None) -> None:
        if not pml4e_index:
            raise Exception('PAE Page tables have no PML4!')
        pdpt = self.read_entries('pdpt', addr)
        for pdpte_index in range(4):
            pdpte = pdpt[pdpte_index]
            if self.is_present(pdpte):
                if self.is_bigpage(pdpte):
                    virt = self.get_virt_addr(0, pdpte_index)
                    phys = pdpte & ADDR_1GB
                    self.print_entry(2, phys, virt, self.get_attr(pdpte))
                else:
                    addr = pdpte & ADDR_4KB
                    self.pt[addr] = 'pd'
                    self.print_entry(2, addr)
                    self.read_pd(addr, 0, pdpte_index)
        return


class c_extended_page_tables(c_4level_page_tables):

    def __init__(self, cs):
        c_4level_page_tables.__init__(self, cs)
        # constants
        self.PHYSICAL_ADDR_NAME = 'GPA'
        self.XWR = {'mask': 0x7, 'offset': 0}
        self.MEM_TYPE = {'mask': 0x7, 'offset': 3}
        self.BIGPAGE = {'mask': 0x1, 'offset': 7}
        self.canonical_msb = 63

    def is_present(self, entry: int) -> bool:
        return self.get_field(entry, self.XWR) != 0

    def is_bigpage(self, entry: int) -> bool:
        return self.get_field(entry, self.BIGPAGE) != 0

    def get_attr(self, entry: int) -> str:
        XWR_DESC = ['---', '--R', '-W-', '-WR', 'X--', 'X-R', 'XW-', 'XWR']
        MEM_DESC = ['UC', 'WC', '02', '03', 'WT', 'WP', 'WB', 'UC-']
        return f'{XWR_DESC[self.get_field(entry, self.XWR)]} {MEM_DESC[self.get_field(entry, self.MEM_TYPE)]}'

    def read_pt_and_show_status(self, path: str, name: str, ptr: int) -> None:
        super(c_extended_page_tables, self).read_pt_and_show_status(path, name, ptr)
        self.check_misconfig(list(self.pt))
        return

    def map_bigpage_1G(self, virt: int, i: int) -> None:
        if self.pointer is None:
            raise Exception('Page Table pointer is undefined!')
        addr = self.pointer
        pml4 = self.read_entries('pml4', addr)
        pml4e = pml4[self.get_field(virt, self.PML4_INDX)]
        if self.is_present(pml4e):
            addr = pml4e & ADDR_4KB
            pdpt = self.read_entries('pdpt', addr)
            new_entry = struct.pack('<Q', ((pdpt[i] | 0x87) & ~ADDR_4KB) | (i << 30))
            self.cs.mem.write_physical_mem(addr + i * 8, 8, new_entry)
        return None


class c_vtd_page_tables(c_extended_page_tables):

    def __init__(self, cs):
        c_extended_page_tables.__init__(self, cs)
        # constants
        self.DID_BUS = {'mask': 0xFF, 'offset': 8}
        self.DID_DEV = {'mask': 0x1F, 'offset': 3}
        self.DID_FUN = {'mask': 0x07, 'offset': 0}
        self.RE_LO_P = {'mask': 0x01, 'offset': 0}
        self.CE_HI_AW = {'mask': 0x07, 'offset': 0}
        self.CE_HI_AVAIL = {'mask': 0x0F, 'offset': 3}
        self.CE_HI_DID = {'mask': 0xFF, 'offset': 8}
        self.CE_LO_P = {'mask': 0x01, 'offset': 0}
        self.CE_LO_FPD = {'mask': 0x01, 'offset': 1}
        self.CE_LO_T = {'mask': 0x03, 'offset': 2}
        # variables
        self.context = {}
        self.domains = {}
        self.cpt = {}

    def read_vtd_context(self, path: str, ptr: int) -> None:
        txt = open(path, 'w')
        try:
            self.out = txt
            addr = ptr & ADDR_4KB
            self.context = {}
            self.domains = {}
            self.cpt = {addr: 'root'}
            self.read_re(addr)

            if len(self.domains) != 0:
                logger().log('[paging] VT-d domains:')
                for domain in sorted(self.domains.keys()):
                    logger().log(f'  0x{domain:016X} ')
            logger().log(f'[paging] Total VT-d domains: {len(self.domains):d}\n')

            logger().log('[paging] VT-d context entries:')
            for source_id in sorted(self.context.keys()):
                self.print_context_entry(source_id, self.context[source_id])

            logger().log('[paging] VT-d context pages:')
            for i in sorted(self.cpt.keys()):
                logger().log(f'    0x{i:013X}  {self.cpt[i]}')
        finally:
            txt.close()
        return

    def read_re(self, addr: int) -> None:
        re = self.read_entries('re', addr, 16)
        for ree_index in range(len(re)):
            ree_lo = re[ree_index][0]
            ree_hi = re[ree_index][1]
            if self.get_field(ree_lo, self.RE_LO_P):
                addr = ree_lo & ADDR_4KB
                self.read_ce(addr, ree_index)
                self.cpt[addr] = 'context'
        return

    def read_ce(self, addr: int, ree_index: int) -> None:
        ce = self.read_entries('ce', addr, 16)
        for cee_index in range(len(ce)):
            cee_lo = ce[cee_index][0]
            cee_hi = ce[cee_index][1]
            if self.get_field(cee_lo, self.CE_LO_P):
                source_id = (ree_index << 8) | cee_index
                self.context[source_id] = [cee_lo, cee_hi]
                if self.get_field(cee_lo, self.CE_LO_T) in (0, 1):
                    slptptr = cee_lo & MAXPHYADDR
                    self.domains[slptptr] = 1
        return

    def print_context_entry(self, source_id: int, cee: Dict[int, int]) -> None:
        if self.get_field(cee[0], self.CE_LO_P):
            info = (
                self.get_field(source_id, self.DID_BUS),
                self.get_field(source_id, self.DID_DEV),
                self.get_field(source_id, self.DID_FUN),
                self.get_field(cee[1], self.CE_HI_DID),
                self.get_field(cee[1], self.CE_HI_AVAIL),
                self.get_field(cee[1], self.CE_HI_AW),
                self.get_field(cee[0], self.CE_LO_T),
                self.get_field(cee[0], self.CE_LO_FPD),
                cee[0] & MAXPHYADDR
            )
            logger().log('  {:02X}:{:02X}.{:X}  DID: {:02X}  AVAIL: {:X}  AW: {:X}  T: {:X}  FPD: {:X}  SLPTPTR: {:016X}'.format(*info))
        return

    def read_page_tables(self, ptr: int) -> None:
        logger().log(f'  Page Tables for domain 0x{ptr:013X}: ')
        super(c_vtd_page_tables, self).read_page_tables(ptr)
        return

    def read_pt_and_show_status(self, path: str, name: str, ptr: int) -> None:
        super(c_vtd_page_tables, self).read_pt_and_show_status(path, name, ptr)
        self.check_misconfig(list(self.cpt))
        return



#!/usr/bin/env python3
# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Setup module to install chipsec package via setuptools
"""

import io
import os
import platform
import logging
import subprocess
from shutil import rmtree
from setuptools import setup, find_packages, Extension, __version__ as _sutver

if _sutver and int(_sutver.split('.')[0]) < 62:
    raise RuntimeError("Setuptools version must be greater than 62.0.0. Please upgrade using 'pip install setuptools --upgrade'")

from setuptools.command.install import install as _install
from setuptools.command.build import build as _build
from setuptools.command.sdist import sdist as _sdist
from setuptools.command.build_ext import build_ext as _build_ext

NO_DRIVER_MARKER_FILE = 'README.NO_KERNEL_DRIVER'


def long_description():
    return open('README').read()


def version():
    return open(os.path.join('chipsec', 'VERSION')).read()


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join(path, filename))
    return paths


skip_driver_opt = [('skip-driver', None,
                    ('Do not build the Chipsec kernel driver. '
                     'Only available on Linux.'))]


class build_ext(_build_ext):
    user_options = _build_ext.user_options + skip_driver_opt
    boolean_options = _build_ext.boolean_options + ['skip-driver']

    def initialize_options(self):
        _build_ext.initialize_options(self)
        self.skip_driver = None

    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Get the value of the skip-driver parameter from the install command.
        self.set_undefined_options('install', ('skip_driver', 'skip_driver'))
        self.set_undefined_options('build', ('skip_driver', 'skip_driver'))

    def _build_linux_driver(self):
        logging.info('Building the linux driver')
        build_driver = os.path.join(self.real_build_lib, 'drivers', 'linux')
        ko_ext = os.path.join(build_driver, 'chipsec.ko')
        # We copy the drivers extension to the build directory.
        self.copy_tree(os.path.join('drivers', 'linux'), build_driver)
        # Run the makefile there.
        subprocess.check_output(['make', '-C', build_driver])
        # And copy the resulting .ko to the right place.
        # That is to the source directory if we are in 'develop' mode,
        # otherwise to the helper subdirectory in the build directory.
        root_dst = '' if self.inplace else self.real_build_lib
        dst = os.path.join(root_dst, 'chipsec', 'helper', 'linux')
        self.copy_file(ko_ext, dst)
        # Finally, we clean up the build directory.
        rmtree(os.path.join(self.real_build_lib, 'drivers'))

    def _build_win_driver(self, path):
        cur_dir = os.getcwd()
        os.chdir(path)
        # Run the makefile there.
        if platform.machine().endswith('64'):
            subprocess.call(['install.cmd'])
        else:
            subprocess.call(['install.cmd', '32'])
        os.chdir(cur_dir)

    def _build_all_win_drivers(self):
        logging.info('Building the windows chipsec driver')
        driver_path = os.path.join('drivers', 'windows', 'chipsec')
        self._build_win_driver(driver_path)
        logging.info('Building the windows pcifilter driver')
        driver_path = os.path.join('drivers', 'windows', 'pcifilter')
        self._build_win_driver(driver_path)

    def run(self):
        # First, we build the standard extensions.
        _build_ext.run(self)

        # Then, we build the compression tools and the driver if required
        def null_builder():
            return None
        driver_build_function = null_builder
        self.real_build_lib = os.path.realpath(self.build_lib)
        if platform.system().lower() == 'linux':
            driver_build_function = self._build_linux_driver
        elif platform.system().lower() == 'windows':
            driver_build_function = self._build_all_win_drivers

        if not self.skip_driver:
            driver_build_function()

    def get_source_files(self):
        files = _build_ext.get_source_files(self)
        if platform.system().lower() == 'linux':
            files.extend(package_files(os.path.join('drivers', 'linux')))
        return files


class install(_install):
    user_options = _install.user_options + skip_driver_opt
    boolean_options = _install.boolean_options + ['skip-driver']

    def initialize_options(self):
        _install.initialize_options(self)
        self.skip_driver = None

        # Do not build the driver if no-driver marker file is present.
        # This marker is only created by an sdist command when
        # 'python setup.py sdist' is executed. This allows having
        # driver-less PIP packages uploaded to PyPi.
        if os.path.exists(NO_DRIVER_MARKER_FILE):
            self.skip_driver = True


class build(_build):
    user_options = _build.user_options + skip_driver_opt
    boolean_options = _build.boolean_options + ['skip-driver']

    def initialize_options(self):
        _build.initialize_options(self)
        self.skip_driver = None


class sdist(_sdist):
    """Build sdist."""

    def make_release_tree(self, base_dir, files):
        pypi_msg = u"""PyPI-distributed chipsec PIP package doesn't contain a pre-built kernel\n""" \
            """driver. Please use it only when a kernel driver is already present on the\n""" \
            """system. Otherwise, please install chipsec from source, using the following\n""" \
            """procedure:""" \
            """\n  https://github.com/chipsec/chipsec/blob/main/chipsec-manual.pdf""" \
            """\n  https://chipsec.github.io/"""
        _sdist.make_release_tree(self, base_dir, files)
        no_driver_marker = os.path.join(base_dir, NO_DRIVER_MARKER_FILE)
        with io.open(no_driver_marker, 'w', encoding='utf-8') as fd:
            fd.write(pypi_msg)


package_data = {
    # Include any configuration file.
    '': ['*.ini', '*.cfg', '*.json'],
    'chipsec': ['*VERSION*', 'WARNING.txt', 'options/*.ini'],
    'chipsec.cfg': ['8086/*.xml', '1022/*.xml', '*.xml', '*.xsd'],
}
data_files = [('', ['chipsec-manual.pdf'])]
install_requires = []
extra_kw = []

if platform.system().lower() == 'windows':
    package_data['chipsec.helper.windows'] = ['windows_amd64/*.sys']
    package_data['chipsec.helper.rwe'] = ['windows_amd64/*.sys']
    package_data['chipsec_tools.compression'] = ['*']
    install_requires.append('pywin32')
    extra_kw = [
        Extension(
            'EfiCompressor',
            sources=[
                os.path.join('chipsec_tools', 'compression', 'Bra86.c'),
                os.path.join('chipsec_tools', 'compression', 'Decompress.c'),
                os.path.join('chipsec_tools', 'compression', 'Compress.c'),
                os.path.join('chipsec_tools', 'compression', 'EfiCompress.c'),
                os.path.join('chipsec_tools', 'compression', 'TianoCompress.c'),
                os.path.join('chipsec_tools', 'compression', 'EfiCompressor.c'),
            ],
            include_dirs=[
                os.path.join('chipsec_tools', 'compression', 'Include'),
                os.path.join('chipsec_tools', 'compression', 'Include', 'Common'),
                os.path.join('chipsec_tools', 'compression', 'Include', 'X64'),
            ],
        )
    ]

elif platform.system().lower() == 'linux':
    package_data['chipsec_tools.compression'] = ['*']
    data_files = [(os.path.abspath(os.path.join(os.sep, 'usr', 'share', 'doc', 'chipsec')), ['chipsec-manual.pdf'])]
    extra_kw = [
        Extension(
            'EfiCompressor',
            sources=[
                os.path.join('chipsec_tools', 'compression', 'Bra86.c'),
                os.path.join('chipsec_tools', 'compression', 'Decompress.c'),
                os.path.join('chipsec_tools', 'compression', 'Compress.c'),
                os.path.join('chipsec_tools', 'compression', 'EfiCompress.c'),
                os.path.join('chipsec_tools', 'compression', 'TianoCompress.c'),
                os.path.join('chipsec_tools', 'compression', 'EfiCompressor.c'),
            ],
            include_dirs=[
                os.path.join('chipsec_tools', 'compression', 'Include'),
                os.path.join('chipsec_tools', 'compression', 'Include', 'Common'),
                os.path.join('chipsec_tools', 'compression', 'Include', 'X64'),
            ],
        )
    ]

setup(
    name='chipsec',
    version=version(),
    description='CHIPSEC: Platform Security Assessment Framework',
    author='CHIPSEC Team',
    author_email='chipsec@intel.com',
    url='https://github.com/chipsec/chipsec',
    download_url='https://github.com/chipsec/chipsec',
    license='GNU General Public License v2 (GPLv2)',
    platforms=['any'],
    long_description=long_description(),

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Security',
        'Topic :: System :: Hardware'
    ],

    data_files=data_files,
    packages=find_packages(exclude=['tests.*', 'tests']),
    package_data=package_data,
    install_requires=install_requires,

    py_modules=['chipsec_main', 'chipsec_util'],
    entry_points={
        'console_scripts': [
            'chipsec_util=chipsec_util:main',
            'chipsec_main=chipsec_main:main',
        ],
    },
    test_suite='tests',
    cmdclass={
        'install': install,
        'build': build,
        'build_ext': build_ext,
        'sdist': sdist,
    },
    ext_modules=extra_kw
)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2024, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Library to build module URLs

"""

import os
import json
import re
from chipsec.library.file import get_main_dir


class url:
    def __init__(self):
        self.url_info = self.get_url_info()
        self.base_url = self.get_base_url()
        self.replace_find = self.url_info.get('replace_find', '')
        self.replace_with = self.url_info.get('replace_with', '')
        self.ends_with = self.url_info.get('ends_with', '')

    def get_url_info(self):
        with open(os.path.join(get_main_dir(), 'chipsec', 'library', 'url_format.json'), 'r') as url_file:
            return json.loads(url_file.read())

    def get_base_url(self):
        if 'base_url' not in self.url_info:
            raise Exception('Missing Base URL in url file')
        return self.url_info['base_url']

    def get_module_url(self, module_name: str) -> str:
        module_name = re.sub(self.replace_find, self.replace_with, module_name)
        module_url = f'{self.base_url}{module_name}{self.ends_with}'
        return module_url



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
>>> chipsec_util cpu info
>>> chipsec_util cpu cr <thread> <cr_number> [value]
>>> chipsec_util cpu cpuid <eax> [ecx]
>>> chipsec_util cpu pt [paging_base_cr3]
>>> chipsec_util cpu topology

Examples:

>>> chipsec_util cpu info
>>> chipsec_util cpu cr 0 0
>>> chipsec_util cpu cr 0 4 0x0
>>> chipsec_util cpu cpuid 0x40000000
>>> chipsec_util cpu pt
>>> chipsec_util cpu topology
"""

from argparse import ArgumentParser

from chipsec.command import BaseCommand, toLoad
from typing import Dict, List, Optional, Union

# ###################################################################
#
# CPU utility
#
# ###################################################################


class CPUCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(usage=__doc__)
        subparsers = parser.add_subparsers()
        parser_info = subparsers.add_parser('info')
        parser_cr = subparsers.add_parser('cr')
        parser_cpuid = subparsers.add_parser('cpuid')
        parser_pt = subparsers.add_parser('pt')
        parser_topology = subparsers.add_parser('topology')
        parser_info.set_defaults(func=self.cpu_info)
        parser_cr.set_defaults(func=self.cpu_cr)
        parser_cpuid.set_defaults(func=self.cpu_cpuid)
        parser_pt.set_defaults(func=self.cpu_pt)
        parser_topology.set_defaults(func=self.cpu_topology)
        parser_cr.add_argument('thread', type=int, nargs='?', default=None)
        parser_cr.add_argument('cr_number', type=int, nargs='?', default=None)
        parser_cr.add_argument('value', type=lambda x: int(x, 0), nargs='?', default=None)
        parser_cpuid.add_argument('eax', type=lambda x: int(x, 0))
        parser_cpuid.add_argument('ecx', type=lambda x: int(x, 0), nargs='?', default=0)
        parser_pt.add_argument('cr3', type=lambda x: int(x, 0), nargs='?', default=None)

        parser.parse_args(self.argv, namespace=CPUCommand)

    def cpu_info(self) -> None:
        self.logger.log("[CHIPSEC] CPU information:")
        ht = self.cs.cpu.is_HT_active()
        threads_per_core = self.cs.cpu.get_number_logical_processor_per_core()
        threads_per_pkg = self.cs.cpu.get_number_logical_processor_per_package()
        cores_per_pkg = self.cs.cpu.get_number_physical_processor_per_package()
        self.logger.log(f'          Hyper-Threading         : {"Enabled" if ht else "Disabled"}')
        self.logger.log(f'          CPU cores per package   : {cores_per_pkg:d}')
        self.logger.log(f'          CPU threads per core    : {threads_per_core:d}')
        self.logger.log(f'          CPU threads per package : {threads_per_pkg:d}')
        try:
            threads_count = self.cs.cpu.get_number_threads_from_APIC_table()
            sockets_count = self.cs.cpu.get_number_sockets_from_APIC_table()
            self.logger.log(f'          Number of sockets       : {sockets_count:d}')
            self.logger.log(f'          Number of CPU threads   : {threads_count:d}')
        except Exception:
            pass

    def cpu_topology(self) -> Dict[str, Dict[int, List[int]]]:
        self.logger.log("[CHIPSEC] CPU information:")
        ht = self.cs.cpu.is_HT_active()
        threads_per_core = self.cs.cpu.get_number_logical_processor_per_core()
        threads_per_pkg = self.cs.cpu.get_number_logical_processor_per_package()
        cores_per_pkg = self.cs.cpu.get_number_physical_processor_per_package()
        num_threads = self.cs.helper.get_threads_count()
        self.logger.log(f'          Hyper-Threading         : {"Enabled" if ht else "Disabled"}')
        self.logger.log(f'          CPU cores per package   : {cores_per_pkg:d}')
        self.logger.log(f'          CPU threads per core    : {threads_per_core:d}')
        self.logger.log(f'          CPU threads per package : {threads_per_pkg:d}')
        self.logger.log(f'          Total threads           : {num_threads:d}')
        topology = self.cs.cpu.get_cpu_topology()
        self.logger.log("          Packages:")
        for p in topology['packages']:
            self.logger.log(f'              {p:d}: {topology["packages"][p]}')
        self.logger.log("          Cores:")
        for c in topology['cores']:
            self.logger.log(f'              {c:d}: {topology["cores"][c]}')

        return topology

    def cpu_cr(self) -> Optional[Union[bool, int]]:
        if self.value is not None:
            self.logger.log(f'[CHIPSEC] CPU{self.thread:d}: write CR{self.cr_number:d} <- 0x{self.value:08X}')
            self.cs.cpu.write_cr(self.thread, self.cr_number, self.value)
            return True
        elif self.cr_number is not None:
            value = self.cs.cpu.read_cr(self.thread, self.cr_number)
            self.logger.log(f'[CHIPSEC] CPU{self.thread:d}: read CR{self.cr_number:d} -> 0x{value:08X}')
            return value
        else:
            for tid in range(self.cs.msr.get_cpu_thread_count()):
                cr0 = self.cs.cpu.read_cr(tid, 0)
                cr2 = self.cs.cpu.read_cr(tid, 2)
                cr3 = self.cs.cpu.read_cr(tid, 3)
                cr4 = self.cs.cpu.read_cr(tid, 4)
                cr8 = self.cs.cpu.read_cr(tid, 8)
                self.logger.log(f'[CHIPSEC][cpu{tid:d}] x86 Control Registers:')
                self.logger.log(f'  CR0: 0x{cr0:016X}')
                self.logger.log(f'  CR2: 0x{cr2:016X}')
                self.logger.log(f'  CR3: 0x{cr3:016X}')
                self.logger.log(f'  CR4: 0x{cr4:016X}')
                self.logger.log(f'  CR8: 0x{cr8:016X}')

    def cpu_cpuid(self) -> None:
        self.logger.log(f'[CHIPSEC] CPUID < EAX: 0x{self.eax:08X}')
        self.logger.log(f'[CHIPSEC]         ECX: 0x{self.ecx:08X}')

        (_eax, _ebx, _ecx, _edx) = self.cs.cpu.cpuid(self.eax, self.ecx)

        self.logger.log("[CHIPSEC] CPUID > EAX: 0x%08X" % _eax)
        self.logger.log("[CHIPSEC]         EBX: 0x%08X" % _ebx)
        self.logger.log("[CHIPSEC]         ECX: 0x%08X" % _ecx)
        self.logger.log("[CHIPSEC]         EDX: 0x%08X" % _edx)

    def cpu_pt(self) -> None:
        if self.cr3 is not None:
            pt_fname = f'pt_{self.cr3:08X}'
            self.logger.log(f'[CHIPSEC] paging physical base (CR3): 0x{self.cr3:016X}')
            self.logger.log(f'[CHIPSEC] dumping paging hierarchy to \'{pt_fname}\'...')
            self.cs.cpu.dump_page_tables(self.cr3, pt_fname)
        else:
            for tid in range(self.cs.msr.get_cpu_thread_count()):
                cr3 = self.cs.cpu.read_cr(tid, 3)
                pt_fname = f'cpu{tid:d}_pt_{cr3:08X}'
                self.logger.log(f'[CHIPSEC][cpu{tid:d}] paging physical base (CR3): 0x{cr3:016X}')
                self.logger.log(f'[CHIPSEC][cpu{tid:d}] dumping paging hierarchy to \'{pt_fname}\'...')
                self.cs.cpu.dump_page_tables(cr3, pt_fname)


commands = {'cpu': CPUCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2018-2019, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

import json
import time
import xml.etree.ElementTree as ET
from typing import Any

import chipsec.library.file
from chipsec.library.logger import logger
from chipsec.library.defines import bytestostring


def get_json_results(json_file: str) -> Any:
    file_data = chipsec.library.file.read_file(json_file)
    if file_data == 0:
        return None
    try:
        json_data = json.loads(bytestostring(file_data))
    except Exception:
        logger().log_error(f'Unable to load JSON file: {json_file}')
        return None
    return json_data


def compute_result_deltas(previous, current):
    deltas = {}
    all_tests = set(current)
    all_tests.update(previous)
    for test in all_tests:
        new_res = prev_res = '-------'
        try:
            new_res = current[test]['result']
        except Exception:
            logger().log_debug('Exception getting current result')
        try:
            prev_res = previous[test]['result']
        except Exception:
            logger().log_debug('Exception getting previous result')
        if new_res != prev_res:
            deltas[test] = {'previous': prev_res, 'current': new_res}
    return deltas


def display_deltas(deltas, hide_time, start_time) -> None:
    logger().log('')
    logger().log('[CHIPSEC] **********************  TEST DELTA SUMMARY  *********************')
    if not hide_time:
        logger().log(f'[CHIPSEC] Time elapsed          {time.time() - start_time:.3f}')
    if deltas:
        logger().log(f'[*] {"Test":46}| {"Previous":10} | {"Current":10}')
        logger().log(f'[*] {"-" * 71}')
        for test in deltas:
            logger().log_bad(f'{test:46}| {deltas[test]["previous"]:10} | {deltas[test]["current"]:10}')
    else:
        logger().log_good('No changes detected.')
    logger().log('[CHIPSEC] *****************************************************************')


def log_deltas_json(deltas, outfile) -> None:
    deltas_json = json.dumps(deltas, sort_keys=True, indent=2, separators=(',', ': '))
    chipsec.library.file.write_file(outfile, deltas_json)


def log_deltas_xml(deltas, outfile) -> None:
    xml_deltas = ET.ElementTree(ET.Element('deltas'))
    delta_root = xml_deltas.getroot()
    delta_root.text = '\n    '
    delta_root.tail = '\n'
    element = None
    for test in deltas:
        element = ET.SubElement(delta_root, 'test', {'current': deltas[test]['current'], 'previous': deltas[test]['previous']})
        element.text = test
        element.tail = '\n    '
    if element is not None:
        element.tail = '\n'
    else:
        delta_root.text = '\n'
    xml_deltas.write(outfile)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2024, AMD Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@amd.com
#


"""
Compatible SMM memory (SMRAM) Protection check module
This CHIPSEC module simply reads SMRAMC and checks that D_LCK is set.

Reference:
https://www.amd.com/en/resources/product-security/bulletin/amd-sb-7014.html

usage:
    ``chipsec_main -m common.smm_close``

Examples:
    >>> chipsec_main.py -m common.smm_close

This module will only run on AMD platforms with SMMMask defined.
"""

from chipsec.module_common import BaseModule, BIOS, SMM
from chipsec.library.returncode import ModuleResult
from typing import List

TAGS = [BIOS, SMM]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'SMM_CLOSE']


class smm_close(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)

    def is_supported(self) -> bool:
        if self.cs.is_amd():
            return self.cs.register.is_defined('SMMMASK')
        self.logger.log("smm_lose is an AMD only module. Skipping module.")
        return False

    def check_SMMClose(self) -> int:
        regval = self.cs.register.read('SMMMASK')
        self.cs.register.print('SMMMASK', regval)
        TSegClose = bool(self.cs.register.get_field('SMMMASK', regval, 'TCLOSE'))
        ASegClose = bool(self.cs.register.get_field('SMMMASK', regval, 'ACLOSE'))

        if (TSegClose or ASegClose):
            self.logger.log_failed(f"TSeg Close {TSegClose}, Aseg Valid {ASegClose}")
            res = ModuleResult.FAILED
        else:
            self.logger.log_passed("ASeg and TSeg are both closed")
            res = ModuleResult.PASSED

        return self.result.getReturnCode(res)

    # --------------------------------------------------------------------------
    # run( module_argv )
    # Required function: run here all tests from this module
    # --------------------------------------------------------------------------
    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test("Verifies AClose and TClose are clear.  Should be clear after SMI Handler.")
        self.res = self.check_SMMClose()

        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from json import loads
import os
from errno import EACCES, EFAULT
from glob import glob
import re
from importlib import import_module
from typing import Any, Dict, List, Optional, Tuple
from chipsec.library.defines import stringtobytes
from chipsec.library.exceptions import OsHelperError
from chipsec.library.file import read_file
from chipsec.library.logger import logger
from chipsec.helper.basehelper import Helper


class ReplayHelper(Helper):

    def __init__(self, filepath: str = ""):
        super(ReplayHelper, self).__init__()
        self.os_system = "test_helper"
        self.os_release = "0"
        self.os_version = "0"
        self.os_machine = "test"
        self.driver_loaded = True
        self.name = "ReplayHelper"
        if filepath and os.path.isfile(filepath):
            self.config_file = filepath
        else:
            files = glob(os.path.join("chipsec", "helper", "record", "*.json"))
            if files:
                self.config_file = max(files, key=os.path.getctime)
            else:
                raise FileNotFoundError("Cannot find a recorded file to load")
        self._data = {}

    def create(self) -> bool:
        return True

    def start(self) -> bool:
        self._load()
        return True

    def stop(self) -> bool:
        return True

    def delete(self) -> bool:
        return True

    def _get_element_eval(self, cmd: str, args: Tuple) -> Optional[Any]:
        element = self._get_element(cmd, args)
        if type(element) is str:
            ematch = re.match(r'^!! <class \'(.*)\'>: (.*)', element)
            if ematch:
                eimport = ematch[1].split('.')
                etype = getattr(import_module('.'.join(eimport[0:-1])), eimport[-1])
                raise etype(ematch[2])
        try:
            evaledobject = eval(element)
        except Exception:
            try:
                evaledobject = stringtobytes(element)
            except Exception:
                return None
        return evaledobject

    def _get_element(self, cmd: str, args: Tuple) -> Optional[Any]:
        try:
            targs = f"({','.join(str(i) for i in args)})"
        except Exception:
            targs = str(args)
        if str(cmd) in self._data:
            if targs in self._data[str(cmd)]:
                try:
                    return self._data[cmd][targs].pop()
                except IndexError as err:
                    logger().log_error(f'Ran out of entries for {str(cmd)} {targs}')
                    err.args = (err.args[0] + f': {str(cmd)} {targs}',)
                    raise err
        logger().log_error(f"Missing entry for {str(cmd)} {targs}")
        return None

    def _load(self) -> None:
        file_data = read_file(self.config_file)
        if file_data == 0:
            raise OsHelperError(f"Unable to open JSON File: {self.config_file}", EACCES)
        try:
            self._data = loads(file_data)
        except Exception:
            raise OsHelperError(f'Unable to load JSON File: {self.config_file}', EFAULT)

    def read_pci_reg(self, bus: int, device: int, function: int, offset: int, size: int) -> int:
        return self._get_element_eval("read_pci_reg", (bus, device, function, offset, size))

    def write_pci_reg(self, bus: int, device: int, function: int, offset: int, value: int, size: int) -> int:
        return self._get_element_eval("write_pci_reg", (bus, device, function, offset, value, size))

    def read_mmio_reg(self, phys_address: int, size: int) -> int:
        return self._get_element_eval("read_mmio_reg", (phys_address, size))

    def write_mmio_reg(self, phys_address: int, size: int, value: int) -> int:
        return self._get_element_eval("write_mmio_reg", (phys_address, size, value))

    #
    # physical_address is 64 bit integer
    #
    def read_phys_mem(self, phys_address: int, length: int) -> bytes:
        return stringtobytes(self._get_element("read_phys_mem", (phys_address, length)))

    def write_phys_mem(self, phys_address: int, length: int, buf: bytes) -> int:
        return self._get_element_eval("write_phys_mem", (phys_address, length, buf))

    def alloc_phys_mem(self, length: int, max_phys_address: int) -> Tuple[int, int]:
        return self._get_element_eval("alloc_phys_mem", (length, max_phys_address))

    def free_phys_mem(self, phys_address: int) -> Optional[int]:
        return self._get_element_eval("free_phys_mem", (phys_address, ))

    def va2pa(self, virtual_address: int) -> Tuple[int, int]:
        return self._get_element_eval("va2pa", (virtual_address, ))

    def map_io_space(self, phys_address: int, length: int, cache_type: int) -> int:
        return self._get_element_eval("map_io_space", (phys_address, length, cache_type))

    #
    # Read/Write I/O port
    #
    def read_io_port(self, io_port: int, size: int) -> int:
        return self._get_element_eval("read_io_port", (io_port, size))

    def write_io_port(self, io_port: int, value: int, size: int) -> int:
        return self._get_element_eval("write_io_port", (io_port, value, size))

    #
    # Read/Write CR registers
    #
    def read_cr(self, cpu_thread_id: int, cr_number: int) -> int:
        return self._get_element_eval("read_cr", (cpu_thread_id, cr_number))

    def write_cr(self, cpu_thread_id: int, cr_number: int, value: int) -> int:
        return self._get_element_eval("write_cr", (cpu_thread_id, cr_number, value))

    #
    # Read/Write MSR on a specific CPU thread
    #
    def read_msr(self, cpu_thread_id: int, msr_addr: int) -> Tuple[int, int]:
        return self._get_element_eval("read_msr", (cpu_thread_id, msr_addr))

    def write_msr(self, cpu_thread_id: int, msr_addr: int, eax: int, edx: int) -> int:
        return self._get_element_eval("write_msr", (cpu_thread_id, msr_addr, eax, edx))

    #
    # Load CPU microcode update on a specific CPU thread
    #
    def load_ucode_update(self, cpu_thread_id: int, ucode_update_buf: bytes) -> bool:
        return self._get_element_eval("load_ucode_update", (cpu_thread_id, ucode_update_buf))

    #
    # Read IDTR/GDTR/LDTR on a specific CPU thread
    #
    def get_descriptor_table(self, cpu_thread_id: int, desc_table_code: int) -> Optional[Tuple[int, int, int]]:
        return self._get_element_eval("get_descriptor_table", (cpu_thread_id, desc_table_code))

    #
    # EFI Variable API
    #
    def EFI_supported(self) -> bool:
        return self._get_element_eval("EFI_supported", ())

    def get_EFI_variable(self, name: str, guid: str) -> Optional[bytes]:
        return self._get_element_eval("get_EFI_variable", (name, guid))

    def set_EFI_variable(self, name: str, guid: str, data: bytes, datasize: Optional[int], attrs: Optional[int]) -> Optional[int]:
        return self._get_element_eval("set_EFI_variable", (name, guid, data, datasize, attrs))

    def delete_EFI_variable(self, name: str, guid: str) -> Optional[int]:
        return self._get_element_eval("delete_EFI_variable", (name, guid))

    def list_EFI_variables(self) -> Optional[Dict[str, List['EfiVariableType']]]:
        return self._get_element_eval("list_EFI_variables", ())

    #
    # ACPI
    #

    def get_ACPI_table(self, table_name: str) -> Optional['Array']:
        return self._get_element_eval("get_ACPI_table", (table_name, ))

    def enum_ACPI_tables(self) -> Optional['Array']:
        return self._get_element_eval("enum_ACPI_tables", ())

    #
    # CPUID
    #
    def cpuid(self, eax: int, ecx: int) -> Tuple[int, int, int, int]:
        return self._get_element_eval("cpuid", (eax, ecx))

    #
    # IOSF Message Bus access
    #
    def msgbus_send_read_message(self, mcr: int, mcrx: int) -> Optional[int]:
        return self._get_element_eval("msgbus_send_read_message", (mcr, mcrx))

    def msgbus_send_write_message(self, mcr: int, mcrx: int, mdr: int) -> None:
        return self._get_element_eval("msgbus_send_write_message", (mcr, mcrx, mdr))

    def msgbus_send_message(self, mcr: int, mcrx: int, mdr: Optional[int]) -> Optional[int]:
        return self._get_element_eval("msgbus_send_message", (mcr, mcrx, mdr))

    #
    # Affinity
    #
    def get_affinity(self) -> Optional[int]:
        return self._get_element_eval("get_affinity", ())

    def set_affinity(self, value: int) -> Optional[int]:
        return self._get_element_eval("set_affinity", (value, ))

    #
    # Logical CPU count
    #
    def get_threads_count(self) -> int:
        return self._get_element_eval("get_threads_count", ())

    #
    # Send SW SMI
    #
    def send_sw_smi(self, cpu_thread_id: int, SMI_code_data: int, _rax: int, _rbx: int, _rcx: int, _rdx: int, _rsi: int, _rdi: int) -> Optional[int]:
        return self._get_element_eval("send_sw_smi", (cpu_thread_id, SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi))

    #
    # Hypercall
    #
    def hypercall(self, rcx: int, rdx: int, r8: int, r9: int, r10: int, r11: int, rax: int, rbx: int, rdi: int, rsi: int, xmm_buffer: int) -> int:
        return self._get_element_eval("hypercall", (rcx, rdx, r8, r9, r10, r11, rax, rbx, rdi, rsi, xmm_buffer))

    #
    # Speculation control
    #
    def retpoline_enabled(self) -> bool:
        return self._get_element_eval("retpoline_enabled", ())


def get_helper():
    return ReplayHelper()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Hyper-V hypercall fuzzer

Usage:
  ``chipsec_main.py -i -m tools.vmm.hv.hypercall -a <mode>[,<vector>,<iterations>] -l log.txt``

    - ``mode``			fuzzing mode

        * ``= status-fuzzing``	finding parameters with hypercall success status
        * ``= params-info``	shows input parameters valid ranges
        * ``= params-fuzzing``	parameters fuzzing based on their valid ranges
        * ``= custom-fuzzing``	fuzzing of known hypercalls
    - ``vector``		hypercall vector
    - ``iterations``		number of hypercall iterations

Note: the fuzzer is incompatible with native VMBus driver (``vmbus.sys``). To use it, remove ``vmbus.sys``
"""
from chipsec.modules.tools.vmm.hv.define import *
from chipsec.modules.tools.vmm.hv.hypercall import *
from chipsec.module_common import BaseModule, CPU

# Hypercall vectors excluded from scan/fuzzing
excluded_hypercalls_from_scan = []
excluded_hypercalls_from_fuzzing = excluded_hypercalls_from_scan + [HV_POST_MESSAGE]

TAGS = [CPU]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'TOOLS', 'VMM', 'HV', 'HYPERCALLFUZZ']


class HypercallFuzz(BaseModule):
    def __init__(self):
        BaseModule.__init__(self)

    def usage(self):
        self.logger.log(__doc__)
        return

    def run(self, module_argv):
        self.logger.start_test("Hyper-V hypercall fuzzer")

        if len(module_argv) > 0:
            command = module_argv[0]
        else:
            self.usage()
            return

        callnum = get_int_arg(module_argv[1]) if len(module_argv) > 1 and module_argv[1] != '' else 'all'
        testnum = get_int_arg(module_argv[2]) if len(module_argv) > 2 and module_argv[2] != '' else 10000000

        hv = HyperVHypercall()
        hv.promt = 'CHIPSEC'

        hv.print_hypervisor_info()

        if hv.hypervisor_present:
            hv.scan_partitionid(range(0x0, 0x100))
            hv.scan_connectionid(range(0x00000, 0x00100) + range(0x10000, 0x10100))

            # Scans for implemented hypercalls and discovers their interface
            hypercalls_for_scanning = list(set(range(0x100)) - set(excluded_hypercalls_from_scan))
            hv.scan_hypercalls(hypercalls_for_scanning)

        if callnum == 'all':
            hypercalls = list(set(hv.hv_hypercalls.keys()) - set(excluded_hypercalls_from_fuzzing))
        else:
            hypercalls = [callnum]

        if command == 'info':
            if hv.hypervisor_present:
                # Print Synthetic MSRs
                hv.print_synthetic_msrs()

                # Print Partition IDs
                hv.print_partitionid()

                # Print Connection IDs
                hv.print_connectionid([])
                hv.print_partition_properties()

                # Print discovered hypercalls and their interface
                hv.print_hypercall_status()

        elif command == 'status-fuzzing':
            for i in hypercalls:
                hv.promt = f'HYPERCALL {i:04X}'
                hv.msg('[*] Scan hypercall for success status')
                hv.scan_for_success_status(i, testnum)

        elif command == 'params-info':
            for i in hypercalls:
                hv.promt = f'HYPERCALL {i:04X}'
                if (hv.hv_hypercalls[i][2] == HV_STATUS_SUCCESS):
                    hv.msg('Scan hypercall for input parameters')
                    hv.scan_input_parameters(i, 32)
                    hv.print_input_parameters(i, 32, [HV_STATUS_SUCCESS])

        elif command == 'params-fuzzing':
            for i in hypercalls:
                hv.promt = f'HYPERCALL {i:04X}'
                if (hv.hv_hypercalls[i][2] == HV_STATUS_SUCCESS):
                    hv.msg('Fuzzing hypercall for input parameters')
                    hv.scan_input_parameters(i, 32)
                    hv.print_input_parameters(i, 32, [HV_STATUS_SUCCESS])
                    hv.input_parameters_fuzzing(i, 32, [HV_STATUS_SUCCESS], testnum)

        elif command == 'custom-fuzzing':
            for i in hypercalls:
                hv.promt = f'HYPERCALL {i:04X}'
                hv.custom_fuzzing(i, testnum)

        else:
            hv.err('Invalid mode!')
            self.usage()

        self.result.setStatusBit(self.result.status.SUCCESS)
        return self.result.getReturnCode(ModuleResult.PASSED)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
CHIPSEC includes functionality for reading and writing the SPI flash. When an image file is created from reading the SPI flash, this image can be parsed to reveal sections, files, variables, etc.

.. warning:: Particular care must be taken when using the SPI write and SPI erase functions. These could make your system unbootable.

A basic forensic operation might be to dump the entire SPI flash to a file. This is accomplished as follows:

``# python chipsec_util.py spi dump rom.bin``

The file rom.bin will contain the full binary of the SPI flash. It can then be parsed using the decode util command.

>>> chipsec_util spi info|dump|read|write|erase|disable-wp [flash_address] [length] [file]

Examples:

>>> chipsec_util spi info
>>> chipsec_util spi dump rom.bin
>>> chipsec_util spi read 0x700000 0x100000 bios.bin
>>> chipsec_util spi write 0x0 flash_descriptor.bin
>>> chipsec_util spi disable-wp
>>> chipsec_util spi sfdp
>>> chipsec_util spi jedec
>>> chipsec_util spi jedec decode
"""

import os
from chipsec.command import BaseCommand, toLoad
from chipsec.hal.spi import SPI, BIOS
from argparse import ArgumentParser


# SPI Flash Controller
class SPICommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util spi', usage=__doc__)
        subparsers = parser.add_subparsers()
        parser_info = subparsers.add_parser('info')
        parser_info.set_defaults(func=self.spi_info)

        parser_dump = subparsers.add_parser('dump')
        parser_dump.add_argument('out_file', type=str, nargs='?', default='rom.bin', help='Output file name [default=rom.bin]')
        parser_dump.set_defaults(func=self.spi_dump)

        parser_read = subparsers.add_parser('read')
        parser_read.add_argument('spi_fla', type=lambda x: int(x, 16), help='Start Address (hex)')
        parser_read.add_argument('length', type=lambda x: int(x, 16), nargs='?', default=0x4, help='Length [default=0x4] (hex)')
        parser_read.add_argument('out_file', type=str, nargs='?', default='read.bin', help='Output file [default=read.bin')
        parser_read.set_defaults(func=self.spi_read)

        parser_write = subparsers.add_parser('write')
        parser_write.add_argument('spi_fla', type=lambda x: int(x, 16), help='Start Address (hex)')
        parser_write.add_argument('filename', type=str, help='File name (hex)')
        parser_write.set_defaults(func=self.spi_write)

        parser_erase = subparsers.add_parser('erase')
        parser_erase.add_argument('spi_fla', type=lambda x: int(x, 16), help='Start Address (hex)')
        parser_erase.set_defaults(func=self.spi_erase)

        parser_disable_wp = subparsers.add_parser('disable-wp')
        parser_disable_wp.set_defaults(func=self.spi_disable_wp)

        parser_sfdp = subparsers.add_parser('sfdp')
        parser_sfdp.set_defaults(func=self.spi_sfdp)

        parser_jedec = subparsers.add_parser('jedec')
        parser_jedec.add_argument('option', type=str, nargs='?', default='', help='Optional decode')
        parser_jedec.set_defaults(func=self.spi_jedec)

        parser.parse_args(self.argv, namespace=self)

    def set_up(self) -> None:
        self._spi = SPI(self.cs)
        self._msg = "it may take a few minutes (use DEBUG or VERBOSE logger options to see progress)"

    def spi_info(self):
        self.logger.log("[CHIPSEC] SPI flash memory information\n")
        self._spi.display_SPI_map()

    def spi_dump(self):
        self.logger.log("[CHIPSEC] Dumping entire SPI flash memory to '{}'".format(self.out_file))
        self.logger.log("[CHIPSEC] {}".format(self._msg))
        # @TODO: don't assume SPI Flash always ends with BIOS region
        (base, limit, _) = self._spi.get_SPI_region(BIOS)
        spi_size = limit + 1
        self.logger.log("[CHIPSEC] BIOS region: base = 0x{:08X}, limit = 0x{:08X}".format(base, limit))
        self.logger.log("[CHIPSEC] Dumping 0x{:08X} bytes (to the end of BIOS region)".format(spi_size))
        buf = self._spi.read_spi_to_file(0, spi_size, self.out_file)
        if buf is None:
            self.logger.log_error("Dumping SPI Flash didn't return any data (turn on VERBOSE)")
        else:
            self.logger.log("[CHIPSEC] Completed SPI flash dump to '{}'".format(self.out_file))

    def spi_read(self):
        self.logger.log("[CHIPSEC] Reading 0x{:x} bytes from SPI Flash starting at FLA = 0x{:X}".format(self.length, self.spi_fla))
        self.logger.log("[CHIPSEC] {}".format(self._msg))
        buf = self._spi.read_spi_to_file(self.spi_fla, self.length, self.out_file)
        if buf is None:
            self.logger.log_error("SPI flash read didn't return any data (turn on VERBOSE)")
        else:
            self.logger.log("[CHIPSEC] Completed SPI flash memory read")

    def spi_write(self):
        if not os.path.exists(self.filename):
            self.logger.log_error("File '{}' doesn't exist".format(self.filename))
            return
        self.logger.log("[CHIPSEC] Writing to SPI flash memory at FLA = 0x{:X} from '{:64s}'".format(self.spi_fla, self.filename))

        if self._spi.write_spi_from_file(self.spi_fla, self.filename):
            self.logger.log("[CHIPSEC] Completed SPI flash memory write")
        else:
            self.logger.log_warning("SPI flash write returned error (turn on VERBOSE)")

    def spi_erase(self):
        self.logger.log("[CHIPSEC] Erasing SPI flash memory block at FLA = 0x{:X}".format(self.spi_fla))

        if self._spi.erase_spi_block(self.spi_fla):
            self.logger.log_good("Completed SPI flash memory erase")
        else:
            self.logger.log_warning("SPI flash erase returned error (turn on VERBOSE)")

    def spi_disable_wp(self):
        self.logger.log("[CHIPSEC] Trying to disable BIOS write protection..")
        #
        # This write protection only matters for BIOS range in SPI flash memory
        #
        if self._spi.disable_BIOS_write_protection():
            self.logger.log_good("BIOS region write protection is disabled in SPI flash")
        else:
            self.logger.log_bad("Couldn't disable BIOS region write protection in SPI flash")

    def spi_sfdp(self):
        self._spi.get_SPI_SFDP()

    def spi_jedec(self):
        if self.option.lower() == 'decode':
            (jedec, man, part) = self._spi.get_SPI_JEDEC_ID_decoded()
            if jedec is not False:
                self.logger.log('    JEDEC ID     : 0x{:06X}'.format(jedec))
                self.logger.log('    Manufacturer : 0x{:02X}     - {}'.format((jedec >> 16) & 0xFF, man))
                self.logger.log('    Device       : 0x{:04X}   - {}'.format(jedec & 0xFFFF, part))
                self.logger.log('')
            else:
                self.logger.log(' JEDEC ID command is not supported')
        else:
            jedec_id = self._spi.get_SPI_JEDEC_ID()
            if jedec_id is not False:
                self.logger.log('    JEDEC ID: 0x{:06X}'.format(jedec_id))
                self.logger.log('')
            else:
                self.logger.log(' JEDEC ID command is not supported')


commands = {'spi': SPICommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2024, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Main functionality to read/write configuration registers based on their XML configuration
"""

from typing import Any, Dict, List, Optional

from chipsec.library.logger import logger
from chipsec.library.options import Options
from chipsec.library.defines import is_all_ones
from chipsec.library.exceptions import CSReadError, RegisterTypeNotFoundError


class RegisterType:
    PCICFG = 'pcicfg'
    MMCFG = 'mmcfg'
    MMIO = 'mmio'
    MSR = 'msr'
    PORTIO = 'io'
    IOBAR = 'iobar'
    MSGBUS = 'msgbus'
    MM_MSGBUS = 'mm_msgbus'
    MEMORY = 'memory'
    IMA = 'indirect'


class Register:
    def __init__(self, cs):
        self.cs = cs
        options = Options()
        self.treat_missing_bar_as_error = options.get_section_data('Library_Config', 'treat_missing_bar_as_error') == "True"

    def is_defined(self, reg_name: str) -> bool:
        """Checks if register is defined in the XML config"""
        try:
            register = self.cs.Cfg.REGISTERS[reg_name]
        except KeyError:
            return False
        if register is not None:
            if 'type' in register and register['type'] in [RegisterType.MMIO, RegisterType.IOBAR]:
                return self._check_bar(register)
            return True
        else:
            return False

    def _check_bar(self, register: Dict[str, Any]) -> bool:
        try:
            if register['type'] == RegisterType.IOBAR:
                return self.cs.iobar.get_IO_BAR_base_address(register['bar'])[0] != 0
            else:
                return self.cs.mmio.get_MMIO_BAR_base_address(register['bar'])[0] != 0
        except CSReadError:
            message = f'BAR {register["bar"]} not found or is invalid'
            if self.treat_missing_bar_as_error:
                logger().log_error(message)
                raise CSReadError(message)
            logger().log_warning(message)
            return False
        
    def is_device_enabled(self, reg_name: str, bus: Optional[int]=None) -> bool:
        """Checks if device is defined in the XML config"""
        if reg_name in self.cs.Cfg.REGISTERS:
            reg = self.get_def(reg_name)
            rtype = reg['type']
            if (rtype == RegisterType.MMCFG) or (rtype == RegisterType.PCICFG):
                if bus is not None:
                    b = bus
                else:
                    b = self.cs.device.get_first_bus(reg)
                d = reg['dev']
                f = reg['fun']
                return self.cs.pci.is_enabled(b, d, f)
            elif (rtype == RegisterType.MMIO):
                bar_name = reg['bar']
                return self.cs.mmio.is_MMIO_BAR_enabled(bar_name, bus)
        return False

    def _get_pci_def(self, reg_def: Dict[str, Any], dev_name: str) -> Dict[str, Any]:
        """Return Bus Dev Fun of a PCI register"""
        if dev_name in self.cs.Cfg.CONFIG_PCI:
            dev = self.cs.Cfg.CONFIG_PCI[dev_name]
            reg_def['bus'] = self.cs.device.get_first_bus(dev)
            reg_def['dev'] = dev['dev']
            reg_def['fun'] = dev['fun']
        return reg_def

    def _get_memory_def(self, reg_def: Dict[str, Any], dev_name: str) -> Dict[str, Any]:
        """Return address access of a MEM register"""
        if dev_name in self.cs.Cfg.MEMORY_RANGES:
            dev = self.cs.Cfg.MEMORY_RANGES[dev_name]
            reg_def['address'] = dev['address']
            reg_def['access'] = dev['access']
        else:
            logger().log_error(f'Memory device {dev_name} not found')
        return reg_def

    def _get_indirect_def(self, reg_def: Dict[str, Any], dev_name: str) -> Dict[str, Any]:
        """Return base index data of a IMA register"""
        if dev_name in self.cs.Cfg.IMA_REGISTERS:
            dev = self.cs.Cfg.IMA_REGISTERS[dev_name]
            if ('base' in dev):
                reg_def['base'] = dev['base']
            else:
                reg_def['base'] = "0"
            if (dev['index'] in self.cs.Cfg.REGISTERS):
                reg_def['index'] = dev['index']
            else:
                logger().log_error(f'Index register {dev["index"]} not found')
            if (dev['data'] in self.cs.Cfg.REGISTERS):
                reg_def['data'] = dev['data']
            else:
                logger().log_error(f'Data register {dev["data"]} not found')
        else:
            logger().log_error(f'Indirect access device {dev_name} not found')
        return reg_def

    def get_def(self, reg_name: str) -> Dict[str, Any]:
        """Return complete register definition"""
        reg_def = self.cs.Cfg.REGISTERS[reg_name]
        if "device" in reg_def:
            dev_name = reg_def["device"]
            if reg_def["type"] in ["pcicfg", "mmcfg"]:
                reg_def = self._get_pci_def(reg_def, dev_name)
            elif reg_def["type"] == "memory":
                reg_def = self._get_memory_def(reg_def, dev_name)
            elif reg_def["type"] == "indirect":
                reg_def = self._get_indirect_def(reg_def, dev_name)
        return reg_def

    def get_bus(self, reg_name: str) -> List[int]:
        """Returns list of buses device/register was discovered on"""
        device = self.cs.Cfg.REGISTERS[reg_name].get('device', '')
        if not device:
            if logger().DEBUG:
                logger().log_important(f"No device found for '{reg_name}'")
            if 'bus' in self.cs.Cfg.REGISTERS[reg_name]:
                return [self.cs.Cfg.REGISTERS[reg_name]['bus']]
            else:
                return []
        return self.cs.device.get_bus(device)

    def _read_pci(self, bus: Optional[int], reg: Dict[str, Any], rtype: str, do_check: bool) -> int:
        """Returns PCI register value"""
        reg_value = 0
        if bus is not None:
            b = self.cs.device.get_first(bus)
        else:
            b = self.cs.device.get_first_bus(reg)
        d = reg['dev']
        f = reg['fun']
        o = reg['offset']
        size = reg['size']
        if do_check and self.cs.consistency_checking:
            if self.cs.pci.get_DIDVID(b, d, f) == (0xFFFF, 0xFFFF):
                raise CSReadError(f'PCI Device is not available ({b}:{d}.{f})')
        if RegisterType.PCICFG == rtype:
            if 1 == size:
                reg_value = self.cs.pci.read_byte(b, d, f, o)
            elif 2 == size:
                reg_value = self.cs.pci.read_word(b, d, f, o)
            elif 4 == size:
                reg_value = self.cs.pci.read_dword(b, d, f, o)
            elif 8 == size:
                reg_value = (self.cs.pci.read_dword(b, d, f, o + 4) << 32) | self.cs.pci.read_dword(b, d, f, o)
        elif RegisterType.MMCFG == rtype:
            reg_value = self.cs.mmio.read_mmcfg_reg(b, d, f, o, size)
        return reg_value

    def _read_mmio(self, bus: Optional[int], reg: Dict[str, Any]) -> int:
        """Returns MMIO register value"""
        _bus = bus
        if self.cs.mmio.get_MMIO_BAR_base_address(reg['bar'], _bus)[0] != 0:
            reg_value = self.cs.mmio.read_MMIO_BAR_reg(reg['bar'], reg['offset'], reg['size'], _bus)
        else:
            raise CSReadError(f'MMIO Bar ({reg["bar"]}) base address is 0')
        return reg_value

    def _read_msr(self, cpu_thread: int, reg: Dict[str, Any]) -> int:
        """Returns MSR register value"""
        (eax, edx) = self.cs.msr.read_msr(cpu_thread, reg['msr'])
        return (edx << 32) | eax

    def _read_portio(self, reg: Dict[str, Any]) -> int:
        """Returns PORTIO register value"""
        port = reg['port']
        size = reg['size']
        return self.cs.io._read_port(port, size)

    def _read_iobar(self, reg: Dict[str, Any]) -> int:
        """Returns IOBAR register value"""
        if self.cs.iobar.get_IO_BAR_base_address(reg['bar'])[0] != 0:
            reg_value = self.cs.iobar.read_IO_BAR_reg(reg['bar'], reg['offset'], reg['size'])
        else:
            raise CSReadError(f'IO Bar ({reg["bar"]}) base address is 0')
        return reg_value

    def _read_memory(self, reg: Dict[str, Any]) -> int:
        """Returns MEM register value"""
        reg_value = 0
        if reg['access'] == 'dram':
            size = reg['size']
            if 1 == size:
                reg_value = self.cs.mem.read_physical_mem_byte(reg['address'])
            elif 2 == size:
                reg_value = self.cs.mem.read_physical_mem_word(reg['address'])
            elif 4 == size:
                reg_value = self.cs.mem.read_physical_mem_dword(reg['address'])
            elif 8 == size:
                reg_value = self.cs.mem.read_physical_mem_qword(reg['address'])
        elif reg['access'] == 'mmio':
            reg_value = self.cs.mmio.read_MMIO_reg(reg['address'], reg['offset'], reg['size'])
        return reg_value

    def _read_ima(self, reg: Dict[str, Any]) -> int:
        """Returns IMA register value"""
        self.write(reg['index'], reg['offset'] + reg['base'])
        return self.read(reg['data'])

    def read(self, reg_name: str, cpu_thread: int = 0, bus: Optional[int] = None, do_check: bool = True) -> int:
        """Reads configuration register (by name)"""

        reg = self.get_def(reg_name)
        rtype = reg['type']
        reg_value = 0
        if (rtype == RegisterType.PCICFG) or (rtype == RegisterType.MMCFG):
            reg_value = self._read_pci(bus, reg, rtype, do_check)
        elif rtype == RegisterType.MMIO:
            reg_value = self._read_mmio(bus, reg)
        elif rtype == RegisterType.MSR:
            reg_value = self._read_msr(cpu_thread, reg)
        elif rtype == RegisterType.PORTIO:
            reg_value = self._read_portio(reg)
        elif rtype == RegisterType.IOBAR:
            reg_value = self._read_iobar(reg)
        elif rtype == RegisterType.MSGBUS:
            reg_value = self.cs.msgbus.msgbus_reg_read(reg['port'], reg['offset'])
        elif rtype == RegisterType.MM_MSGBUS:
            reg_value = self.cs.msgbus.mm_msgbus_reg_read(reg['port'], reg['offset'])
        elif rtype == RegisterType.MEMORY:
            reg_value = self._read_memory(reg)
        elif rtype == RegisterType.IMA:
            reg_value = self._read_ima(reg)
        else:
            raise RegisterTypeNotFoundError(f'Register type not found: {rtype}')

        return reg_value

    def read_all(self, reg_name: str, cpu_thread: int = 0) -> List[int]:
        """Reads all configuration register instances (by name)"""
        values = []
        bus_data = self.get_bus(reg_name)
        reg = self.get_def(reg_name)
        rtype = reg['type']
        if RegisterType.MSR == rtype:
            topology = self.cs.cpu.get_cpu_topology()
            if 'scope' in reg.keys() and reg['scope'] == 'packages':
                packages = topology['packages']
                threads_to_use = [packages[p][0] for p in packages]
            elif 'scope' in reg.keys() and reg['scope'] == 'cores':
                cores = topology['cores']
                threads_to_use = [cores[p][0] for p in cores]
            else:  # Default to threads
                threads_to_use = range(self.cs.helper.get_threads_count())
            for t in threads_to_use:
                values.append(self.read(reg_name, t))
        elif rtype in [RegisterType.MMCFG, RegisterType.PCICFG, RegisterType.MMIO]:
            if bus_data:
                for bus in bus_data:
                    values.append(self.read(reg_name, cpu_thread, bus))
        else:
            values.append(self.read(reg_name, cpu_thread))
        return values

    def _write_pci(self, bus: Optional[int], reg: Dict[str, Any], rtype: str, reg_value: int) -> None:
        """Writes PCI register value"""
        if bus is not None:
            b = bus
        else:
            b = self.cs.device.get_first_bus(reg)
        d = reg['dev']
        f = reg['fun']
        o = reg['offset']
        size = reg['size']
        if RegisterType.PCICFG == rtype:
            if 1 == size:
                self.cs.pci.write_byte(b, d, f, o, reg_value)
            elif 2 == size:
                self.cs.pci.write_word(b, d, f, o, reg_value)
            elif 4 == size:
                self.cs.pci.write_dword(b, d, f, o, reg_value)
            elif 8 == size:
                self.cs.pci.write_dword(b, d, f, o, (reg_value & 0xFFFFFFFF))
                self.cs.pci.write_dword(b, d, f, o + 4, (reg_value >> 32 & 0xFFFFFFFF))
        elif RegisterType.MMCFG == rtype:
            self.cs.mmio.write_mmcfg_reg(b, d, f, o, size, reg_value)

    def _write_msr(self, reg: Dict[str, Any], reg_value: int, cpu_thread: int) -> None:
        """Writes MSR register value"""
        eax = (reg_value & 0xFFFFFFFF)
        edx = ((reg_value >> 32) & 0xFFFFFFFF)
        self.cs.msr.write_msr(cpu_thread, reg['msr'], eax, edx)

    def _write_portio(self, reg: Dict[str, Any], reg_value: int) -> None:
        """Writes PORTIO register value"""
        port = reg['port']
        size = reg['size']
        self.cs.io._write_port(port, reg_value, size)

    def _write_memory(self, reg: Dict[str, Any], reg_value: int) -> None:
        """Writes MEM register value"""
        if reg['access'] == 'dram':
            self.cs.mem.write_physical_mem(reg['address'], reg['size'], reg_value)
        elif reg['access'] == 'mmio':
            self.cs.mmio.write_MMIO_reg(reg['address'], reg['offset'], reg_value, reg['size'])

    def _write_ima(self, reg: Dict[str, Any], reg_value: int) -> None:
        """Writes IMA register value"""
        self.write(reg['index'], reg['offset'] + reg['base'])
        self.write(reg['data'], reg_value)

    def write(self, reg_name: str, reg_value: int, cpu_thread: int = 0, bus: Optional[int] = None) -> bool:
        """Writes configuration register (by name)"""
        reg = self.get_def(reg_name)
        rtype = reg['type']
        if (rtype == RegisterType.PCICFG) or (rtype == RegisterType.MMCFG):
            self._write_pci(bus, reg, rtype, reg_value)
        elif rtype == RegisterType.MMIO:
            self.cs.mmio.write_MMIO_BAR_reg(reg['bar'], reg['offset'], reg_value, reg['size'], bus)
        elif rtype == RegisterType.MSR:
            self._write_msr(reg, reg_value, cpu_thread)
        elif rtype == RegisterType.PORTIO:
            self._write_portio(reg, reg_value)
        elif rtype == RegisterType.IOBAR:
            self.cs.iobar.write_IO_BAR_reg(reg['bar'], reg['offset'], reg['size'], reg_value)
        elif rtype == RegisterType.MSGBUS:
            self.cs.msgbus.msgbus_reg_write(reg['port'], reg['offset'], reg_value)
        elif rtype == RegisterType.MM_MSGBUS:
            self.cs.msgbus.mm_msgbus_reg_write(reg['port'], reg['offset'], reg_value)
        elif rtype == RegisterType.MEMORY:
            self._write_memory(reg, reg_value)
        elif rtype == RegisterType.IMA:
            self._write_ima(reg, reg_value)
        else:
            raise RegisterTypeNotFoundError(f'Register type not found: {rtype}')
        return True

    def _write_msr_all(self, reg: Dict[str, Any], reg_name: str, reg_values: List[int]) -> bool:
        """Writes values to all instances of an MSR register"""
        ret = False
        topology = self.cs.cpu.get_cpu_topology()
        if 'scope' in reg.keys() and reg['scope'] == "packages":
            packages = topology['packages']
            threads_to_use = [packages[p][0] for p in packages]
        elif 'scope' in reg.keys() and reg['scope'] == "cores":
            cores = topology['cores']
            threads_to_use = [cores[p][0] for p in cores]
        else:  # Default to threads
            threads_to_use = range(self.cs.helper.get_threads_count())
        if len(reg_values) == len(threads_to_use):
            value = 0
            for t in threads_to_use:
                self.write(reg_name, reg_values[value], t)
                value += 1
            ret = True
        return ret

    def _write_pci_all(self, reg_name: str, reg_values: List[int], cpu_thread: int, bus_data: List[int]) -> bool:
        """Writes values to all instances of a PCI register"""
        ret = False
        values = len(bus_data)
        if len(reg_values) == values:
            for index in range(values):
                self.write(reg_name, reg_values[index], cpu_thread, bus_data[index])
            ret = True
        return ret

    def write_all(self, reg_name: str, reg_values: List[int], cpu_thread: int = 0) -> bool:
        """Writes all configuration register instances (by name)"""
        reg = self.get_def(reg_name)
        rtype = reg['type']
        bus_data = self.get_bus(reg_name)
        ret = False
        if rtype == RegisterType.MSR:
            ret = self._write_msr_all(reg, reg_name, reg_values)
        elif rtype in [RegisterType.MMCFG, RegisterType.PCICFG, RegisterType.MMIO] and bus_data:
            ret = self._write_pci_all(reg_name, reg_values, cpu_thread, bus_data)
        else:
            if len(reg_values) == 1:
                ret = self.write(reg_name, reg_values[0])
        if not ret and logger().DEBUG:
            logger().log("[write_register_all] There is a mismatch in the number of register values and registers to write")
        return ret

    def write_all_single(self, reg_name: str, reg_value: int, cpu_thread: int = 0) -> bool:
        """Writes all configuration register instances (by name)"""
        reg = self.get_def(reg_name)
        rtype = reg['type']
        bus_data = self.get_bus(reg_name)
        if RegisterType.MSR == rtype:
            topology = self.cs.cpu.get_cpu_topology()
            if 'scope' in reg.keys() and reg['scope'] == "packages":
                packages = topology['packages']
                threads_to_use = [packages[p][0] for p in packages]
            elif 'scope' in reg.keys() and reg['scope'] == "cores":
                cores = topology['cores']
                threads_to_use = [cores[p][0] for p in cores]
            else:  # Default to threads
                threads_to_use = range(self.cs.helper.get_threads_count())
            for t in threads_to_use:
                self.write(reg_name, reg_value, t)
        elif rtype in [RegisterType.MMCFG, RegisterType.PCICFG, RegisterType.MMIO] and bus_data:
            for bus in bus_data:
                self.write(reg_name, reg_value, cpu_thread, bus)
        else:
            self.write(reg_name, reg_value)
        return True

    def read_dict(self, reg_name: str) -> Dict[str, Any]:
        """Returns complete register definition (with values)"""
        reg_value = self.read(reg_name)
        reg_def = self.get_def(reg_name)
        result = reg_def
        result['value'] = reg_value
        for f in reg_def['FIELDS']:
            result['FIELDS'][f]['bit'] = field_bit = int(reg_def['FIELDS'][f]['bit'])
            result['FIELDS'][f]['size'] = field_size = int(reg_def['FIELDS'][f]['size'])
            field_mask = 0
            for i in range(field_size):  #TODO: update this routine
                field_mask = (field_mask << 1) | 1
            result['FIELDS'][f]['value'] = (reg_value >> field_bit) & field_mask
        return result

    def get_field_mask(self, reg_name: str, reg_field: Optional[str] = None, preserve_field_position: bool = False) -> int:
        """Returns the field mask for a register field definition (by name)"""
        reg_def = self.get_def(reg_name)
        if reg_field is not None:
            field_attrs = reg_def['FIELDS'][reg_field]
            mask_start = int(field_attrs['bit'])
            mask = (1 << int(field_attrs['size'])) - 1
        else:
            mask_start = 0
            mask = (1 << (reg_def['size'] * 8)) - 1
        if preserve_field_position:
            return mask << mask_start
        else:
            return mask

    def get_field(self, reg_name: str, reg_value: int, field_name: str, preserve_field_position: bool = False) -> int:
        """Reads the value of the field (by name) of configuration register (by register value)"""
        field_attrs = self.get_def(reg_name)['FIELDS'][field_name]
        field_bit = int(field_attrs['bit'])
        field_mask = (1 << int(field_attrs['size'])) - 1
        if preserve_field_position:
            return reg_value & (field_mask << field_bit)
        else:
            return (reg_value >> field_bit) & field_mask

    def get_field_all(self, reg_name: str, reg_values: List[int], field_name: str, preserve_field_position: bool = False) -> List[int]:
        """Reads the value of the field (by name) of all configuration register instances (by register value)"""
        values = []
        for reg_value in reg_values:
            values.append(self.get_field(reg_name, reg_value, field_name, preserve_field_position))
        return values

    def set_field(self, reg_name: str, reg_value: int, field_name: str, field_value: int, preserve_field_position: bool = False) -> int:
        """writes the value of the field (by name) of configuration register (by register value)"""
        field_attrs = self.get_def(reg_name)['FIELDS'][field_name]
        field_bit = int(field_attrs['bit'])
        field_mask = (1 << int(field_attrs['size'])) - 1
        reg_value &= ~(field_mask << field_bit)  # keep other fields
        if preserve_field_position:
            reg_value |= (field_value & (field_mask << field_bit))
        else:
            reg_value |= ((field_value & field_mask) << field_bit)
        return reg_value

    def set_field_all(self, reg_name: str, reg_values: List[int], field_name: str, field_value: int, preserve_field_position: bool = False) -> List[int]:
        """Writes the value of the field (by name) of all configuration register instances (by register value)"""
        values = []
        for reg_value in reg_values:
            values.append(self.set_field(reg_name, reg_value, field_name, field_value, preserve_field_position))
        return values

    def read_field(self, reg_name: str, field_name: str, preserve_field_position: bool = False, cpu_thread: int = 0, bus: Optional[int] = None) -> int:
        """Reads the value of the field (by name) of configuration register (by register name)"""
        reg_value = self.read(reg_name, cpu_thread, bus)
        return self.get_field(reg_name, reg_value, field_name, preserve_field_position)

    def read_field_all(self, reg_name: str, field_name: str, preserve_field_position: bool = False, cpu_thread: int = 0) -> List[int]:
        """Reads the value of the field (by name) of all configuration register instances (by register name)"""
        reg_values = self.read_all(reg_name, cpu_thread)
        return self.get_field_all(reg_name, reg_values, field_name, preserve_field_position)

    def write_field(self, reg_name: str, field_name: str, field_value: int, preserve_field_position: bool = False, cpu_thread: int = 0) -> bool:
        """Writes the value of the field (by name) of configuration register (by register name)"""
        try:
            reg_value = self.read(reg_name, cpu_thread)
            reg_value_new = self.set_field(reg_name, reg_value, field_name, field_value, preserve_field_position)
            ret = self.write(reg_name, reg_value_new, cpu_thread)
        except Exception:
            ret = False
        return ret

    def write_field_all(self, reg_name: str, field_name: str, field_value: int, preserve_field_position: bool = False, cpu_thread: int = 0) -> bool:
        """Writes the value of the field (by name) of all configuration register instances (by register name)"""
        reg_values = self.read_all(reg_name, cpu_thread)
        reg_values_new = self.set_field_all(reg_name, reg_values, field_name, field_value, preserve_field_position)
        return self.write_all(reg_name, reg_values_new, cpu_thread)

    def has_field(self, reg_name: str, field_name: str) -> bool:
        """Checks if the register has specific field"""
        try:
            reg_def = self.get_def(reg_name)
        except KeyError:
            return False
        if 'FIELDS' not in reg_def:
            return False
        return (field_name in reg_def['FIELDS'])

    def has_all_fields(self, reg_name: str, field_list: List[str]) -> bool:
        """Checks if the register as all fields specified in list"""
        ret = True
        for field in field_list:
            ret = ret and self.has_field(reg_name, field)
            if not ret:
                break
        return ret

    def _fields_str(self, reg_def: Dict[str, Any], reg_val: int) -> str:
        """Returns string of all fields of a register and their values."""
        reg_fields_str = ''
        if 'FIELDS' in reg_def:
            reg_fields_str += '\n'
            # sort fields by their bit position in the register
            sorted_fields = sorted(reg_def['FIELDS'].items(), key=lambda field: int(field[1]['bit']))
            for f in sorted_fields:
                field_attrs = f[1]
                field_bit = int(field_attrs['bit'])
                field_size = int(field_attrs['size'])
                field_mask = 0
                for i in range(field_size):
                    field_mask = (field_mask << 1) | 1
                field_value = (reg_val >> field_bit) & field_mask
                field_desc = f' << {field_attrs["desc"]} ' if (field_attrs['desc'] != '') else ''
                reg_fields_str += f'    [{field_bit:02d}] {f[0]:16} = {field_value:X}{field_desc}\n'

        if reg_fields_str:
            reg_fields_str = reg_fields_str[:-1]
        return reg_fields_str

    def print(self, reg_name: str, reg_val: int, bus: Optional[int] = None, cpu_thread: int = 0) -> str:
        """Prints configuration register"""
        reg = self.get_def(reg_name)
        rtype = reg['type']
        reg_str = ''
        reg_width = reg["size"] * 2
        reg_val_str = f'0x{reg_val:0{reg_width:d}X}'
        if RegisterType.PCICFG == rtype or RegisterType.MMCFG == rtype:
            if bus is not None:
                b = bus
            else:
                b = self.cs.device.get_first_bus(reg)
            d = reg['dev']
            f = reg['fun']
            o = reg['offset']
            mmcfg_off_str = ''
            if RegisterType.MMCFG == rtype:
                mmcfg_off = (b * 32 * 8 + d * 8 + f) * 0x1000 + o
                mmcfg_off_str += f', MMCFG + 0x{mmcfg_off:X}'
            reg_str = f'[*] {reg_name} = {reg_val_str} << {reg["desc"]} (b:d.f {b:02d}:{d:02d}.{f:d} + 0x{o:X}{mmcfg_off_str})'
        elif RegisterType.MMIO == rtype:
            reg_str = f'[*] {reg_name} = {reg_val_str} << {reg["desc"]} ({reg["bar"]} + 0x{reg["offset"]:X})'
        elif RegisterType.MSR == rtype:
            reg_str = f'[*] {reg_name} = {reg_val_str} << {reg["desc"]} (MSR 0x{reg["msr"]:X} Thread 0x{cpu_thread:X})'
        elif RegisterType.PORTIO == rtype:
            reg_str = f'[*] {reg_name} = {reg_val_str} << {reg["desc"]} (I/O port 0x{reg["port"]:X})'
        elif RegisterType.IOBAR == rtype:
            reg_str = f'[*] {reg_name} = {reg_val_str} << {reg["desc"]} (I/O {reg["bar"]} + 0x{reg["offset"]:X})'
        elif RegisterType.MSGBUS == rtype or RegisterType.MM_MSGBUS == rtype:
            reg_str = f'[*] {reg_name} = {reg_val_str} << {reg["desc"]} (msgbus port 0x{reg["port"]:X}, off 0x{reg["offset"]:X})'
        elif RegisterType.IMA == rtype:
            reg_str = f'[*] {reg_name} = {reg_val_str} << {reg["desc"]} (indirect access via {reg["index"]}/{reg["data"]}, base 0x{reg["base"]:X}, off 0x{reg["offset"]:X})'
        else:
            reg_str = f'[*] {reg_name} = {reg_val_str} << {reg["desc"]}'

        reg_str += self._fields_str(reg, reg_val)
        logger().log(reg_str)
        return reg_str

    def print_all(self, reg_name: str, cpu_thread: int = 0) -> str:
        """Prints all configuration register instances"""
        reg_str = ''
        bus_data = self.get_bus(reg_name)
        reg = self.get_def(reg_name)
        rtype = reg['type']
        if RegisterType.MSR == rtype:
            topology = self.cs.cpu.get_cpu_topology()
            if 'scope' in reg.keys() and reg['scope'] == "packages":
                packages = topology['packages']
                threads_to_use = [packages[p][0] for p in packages]
            elif 'scope' in reg.keys() and reg['scope'] == "cores":
                cores = topology['cores']
                threads_to_use = [cores[p][0] for p in cores]
            else:
                threads_to_use = range(self.cs.helper.get_threads_count())
            for t in threads_to_use:
                reg_val = self.read(reg_name, t)
                reg_str += self.print(reg_name, reg_val, cpu_thread=t)
        elif rtype in [RegisterType.MMCFG, RegisterType.PCICFG, RegisterType.MMIO] and bus_data:
            for bus in bus_data:
                reg_val = self.read(reg_name, cpu_thread, bus)
                reg_str += self.print(reg_name, reg_val, bus)
        else:
            reg_val = self.read(reg_name, cpu_thread)
            reg_str = self.print(reg_name, reg_val)
        return reg_str

    def is_msr(self, reg_name: str) -> bool:
        """Returns True if register is type `msr`"""
        if self.is_defined(reg_name):
            if self.cs.Cfg.REGISTERS[reg_name]['type'].lower() == 'msr':
                return True
        return False

    def is_pci(self, reg_name: str) -> bool:
        """Returns True if register is type `pcicfg` or `mmcfg`"""
        if self.is_defined(reg_name):
            reg_def = self.cs.Cfg.REGISTERS[reg_name]
            if (reg_def['type'].lower() == 'pcicfg') or (reg_def['type'].lower() == 'mmcfg'):
                return True
        return False

    def is_all_ffs(self, reg_name: str, value: int) -> bool:
        """Returns True if register value is all 0xFFs"""
        if self.is_msr(reg_name):
            size = 8
        else:
            size = self.get_def(reg_name)['size']
        return is_all_ones(value, size)

    def is_field_all_ones(self, reg_name: str, field_name: str, value: int) -> bool:
        """Returns True if field value is all ones"""
        reg_def = self.get_def(reg_name)
        size = reg_def['FIELDS'][field_name]['size']
        return is_all_ones(value, size, 1)
